# coding: utf-8
from __future__ import print_function, division, absolute_import

from atropos.seqio import Sequence
from atropos.modifiers import *
from atropos.adapters import *

DUMMY_ADAPTER = Adapter("ACGT", FRONT)
def front_match(read):
	match = Match(0, 2, 0, 2, 1, 0, True, DUMMY_ADAPTER, read)
	match_info = MatchInfo("read", 0, 0, 2, "", "AC", "ACGTAC", "adapter", "", "##", "######", True, 2, 2, 2)
	return (match, [match_info])
def back_match(read):
	match = Match(6, 8, 6, 8, 1, 0, False, DUMMY_ADAPTER, read)
	match_info = MatchInfo("read", 0, 6, 8, "ACGTAC", "GT", "", "adapter", "######", "##", "", False, 2, 2, 2)
	return (match, [match_info])
	
def test_unconditional_cutter():
	uc = UnconditionalCutter(lengths=[5])
	s = Sequence("read1", 'abcdefg')
	assert UnconditionalCutter(lengths=[2])(s).sequence == 'cdefg'
	assert UnconditionalCutter(lengths=[-2])(s).sequence == 'abcde'
	assert UnconditionalCutter(lengths=[100])(s).sequence == ''
	assert UnconditionalCutter(lengths=[-100])(s).sequence == ''


def test_nend_trimmer():
	trimmer = NEndTrimmer()
	seqs = ['NNNNAAACCTTGGNNN', 'NNNNAAACNNNCTTGGNNN', 'NNNNNN']
	trims = ['AAACCTTGG', 'AAACNNNCTTGG', '']
	for seq, trimmed in zip(seqs, trims):
		_seq = Sequence('read1', seq, qualities='#'*len(seq))
		_trimmed = Sequence('read1', trimmed, qualities='#'*len(trimmed))
		assert trimmer(_seq) == _trimmed


def test_quality_trimmer():
	read = Sequence('read1', 'ACGTTTACGTA', '##456789###')

	qt = QualityTrimmer(10, 10, 33)
	assert qt(read) == Sequence('read1', 'GTTTAC', '456789')

	qt = QualityTrimmer(0, 10, 33)
	assert qt(read) == Sequence('read1', 'ACGTTTAC', '##456789')

	qt = QualityTrimmer(10, 0, 33)
	assert qt(read) == Sequence('read1', 'GTTTACGTA', '456789###')

def test_Modifiers_single():
	m = Modifiers(paired=False)
	m.add_modifier(UnconditionalCutter, lengths=[5])
	assert len(m.modifiers1) == 1
	assert list(m.modifiers1.keys())[0] == UnconditionalCutter
	assert list(m.modifiers1.values())[0].__class__ == UnconditionalCutter
	assert len(m.modifiers2) == 0
	# test single-end
	read = Sequence('read1', 'ACGTTTACGTA', '##456789###')
	mod_read, mod_bp = m.modify(read)
	assert mod_read[0].sequence == 'TACGTA'

def test_Modifiers_paired_legacy():
	m = Modifiers(paired="first")
	m.add_modifier(UnconditionalCutter, lengths=[5])
	read = Sequence('read1', 'ACGTTTACGTA', '##456789###')
	read2 = Sequence('read1', 'ACGTTTACGTA', '##456789###')
	(mod_read, mod_read2), mod_bp = m.modify((read, read2))
	assert mod_read.sequence == 'TACGTA'
	assert mod_read2.sequence == 'ACGTTTACGTA'

def test_Modifiers_paired_both():
	m = Modifiers(paired="both")
	m.add_modifier(UnconditionalCutter, read=1|2, lengths=[5])
	assert len(m.modifiers1) == 1
	assert len(m.modifiers2) == 1
	assert list(m.modifiers1.keys())[0] == list(m.modifiers2.keys())[0] == UnconditionalCutter
	assert list(m.modifiers1.values())[0].__class__ == list(m.modifiers2.values())[0].__class__  == UnconditionalCutter
	read = Sequence('read1', 'ACGTTTACGTA', '##456789###')
	read2 = Sequence('read1', 'ACGTTTACGTA', '##456789###')
	(mod_read, mod_read2), mod_bp = m.modify((read, read2))
	assert mod_read.sequence == 'TACGTA'
	assert mod_read2.sequence == 'TACGTA'

def test_min_cutter_T_T():
	unconditional_before = UnconditionalCutter((2,-2))
	unconditional_after = UnconditionalCutter((1,-1))
	min_trimmer = MinCutter((5,-5), True, True)
	
	read1 = Sequence('read1', "CAATCGATCGAACGTACCGAT")
	assert read1.clipped == [0,0,0,0], str(read1.clipped)
	read1 = unconditional_before(read1)
	assert read1.sequence == "ATCGATCGAACGTACCG"
	assert read1.clipped == [2,2,0,0], str(read1.clipped)
	
	# test without adapter trimming
	assert min_trimmer(read1).sequence == "ATCGATCGAACGTACCG"
	
	# test with adapter trimming
	read2 = read1[:]
	read2.sequence = "ATCGAACGTACCG"
	read2.match, read2.match_info = front_match(read2)
	read3 = min_trimmer(read2)
	assert read3.sequence == "TCGAACGTACCG", read3.sequence
	assert read3.clipped == [2,2,1,0]
	
	# test with subsequent clipping
	read4 = unconditional_after(read2)
	assert read4.sequence == "TCGAACGTACC", read4.sequence
	assert read4.clipped == [2,2,1,1], read4.clipped
	read5 = min_trimmer(read4)
	assert read5.sequence == "TCGAACGTACC", read5.sequence
	assert read5.clipped == [2,2,1,1], read5.clipped

def test_min_cutter_F_T():
	unconditional_before = UnconditionalCutter((2,-2))
	unconditional_after = UnconditionalCutter((1,-1))
	min_trimmer = MinCutter((5,-5), False, True)
	
	read1 = Sequence('read1', "CAATCGATCGAACGTACCGAT")
	read1 = unconditional_before(read1)
	assert read1.sequence == "ATCGATCGAACGTACCG"
	assert read1.clipped == [2,2,0,0]
	
	# test without adapter trimming
	assert min_trimmer(read1).sequence == "ATCGATCGAACGTACCG"
	
	# test with adapter trimming
	read2 = read1[:]
	read2.match, read2.match_info = front_match(read2)
	read2.sequence = "CGATCGAACGTACCG"
	read3 = min_trimmer(read2)
	assert read3.sequence == "GAACGTACCG", read3.sequence
	assert read3.clipped == [2,2,5,0]
	
	# test with subsequent clipping
	read4 = unconditional_after(read2)
	assert read4.sequence == "GATCGAACGTACC"
	assert read4.clipped == [2,2,1,1]
	read5 = min_trimmer(read4)
	assert read5.sequence == "GAACGTACC", read5.sequence
	assert read5.clipped == [2,2,5,1]

def test_min_cutter_T_F():
	unconditional_before = UnconditionalCutter((2,-2))
	min_trimmer = MinCutter((4,-4), True, False)
	
	read1 = Sequence('read1', "CAATCGATCGAACGTACCGAT")
	read1 = unconditional_before(read1)
	assert read1.sequence == "ATCGATCGAACGTACCG"
	assert read1.clipped == [2,2,0,0]
	
	# test without adapter trimming
	assert min_trimmer(read1).sequence == "CGATCGAACGTAC"
	
def test_non_directional_bisulfite_trimmer():
	trimmer = NonDirectionalBisulfiteTrimmer()
	read1 = Sequence('read1', "CAATCGATCGA")
	read2 = Sequence('read2', "CTATCGATC")
	read2.match, read2.match_info = back_match(read2)
	read3 = Sequence('read3', "CTATCGATCCA")
	#assert trimmer(read1).sequence == "ATCGATC"
	assert trimmer(read2).sequence == "CTATCGA"
	assert trimmer(read3).sequence == "CTATCGATCCA"

def test_EpiGnome_trimmer():
	trimmer = EpiGnomeBisulfiteTrimmer()
	read1 = Sequence('read1', "CTATCGATCCACGAGACTAAC")
	assert trimmer(read1).sequence == "ATCCACGAGACTAAC"

def test_Swift_trimmer():
	trimmer = SwiftBisulfiteTrimmer()
	short_read = Sequence('read1', "ACGTACGTACGTACGT")
	long_seq = "".join(["ACGT"] * 30)
	long_read = Sequence('read2', long_seq)
	assert trimmer(short_read).sequence == "GTACGT"
	assert trimmer(long_read).sequence == long_seq[10:-10]

def test_overlapping():
	trimmer = MergeOverlapping(min_overlap=10, error_rate = 0.1)
	a1 = 'AGATCGGAAGAGCGTCGTGTAGGGAAAGAGTGTAGATCTC'
	a2 = 'AGATCGGAAGAGCACACGTCTGAACTCCAGTCACGAGTTA'
	frag = 'CCAAGCAGACATTCACTCAGATTGCA'
	r1 = (frag + a1)[0:40]
	q1 = '#' * 40
	r2 = reverse_complement(a2 + frag)[0:40]
	q2 = '!' * 40
	parser = AdapterParser()
	adapter1 = parser.parse(a1)
	adapter2 = parser.parse(reverse_complement(a2))
	cutter = AdapterCutter([adapter1, adapter2])
	read1 = Sequence('foo', r1, q1)
	read1 = cutter(read1)
	assert len(read1) == 26
	read2 = Sequence('foo', r2, q2)
	read2 = cutter(read2)
	assert len(read2) == 26
	
	# complete overlap
	read1_merged, read2_merged = trimmer(read1, read2)
	assert read1_merged.merged
	assert read2_merged is None
	assert read1 == read1_merged
	
	# partial overlap
	read1.merged = False
	read2 = read2.subseq(0, 24)[2]
	read1_merged, read2_merged = trimmer(read1, read2)
	assert read1_merged.merged
	assert read2_merged is None
	assert read1 == read1_merged
	
	# partial overlap r1, r2
	read1.merged = False
	read1 = read1.subseq(0, 24)[2]
	read1_merged, read2_merged = trimmer(read1, read2)
	assert read1_merged.merged
	assert read2_merged is None
	assert len(read1_merged) == 26
	assert read1_merged.sequence == 'CCAAGCAGACATTCACTCAGATTGCA'
	assert read1_merged.qualities == ('#' * 24) + ( '!' * 2)
	
	# errors
	# round(0.1 * 24) = 2, so 2 errors should pass but 3 should not
	read1.merged = False
	r1_seq = list(read1.sequence)
	r1_seq[10] = reverse_complement(r1_seq[10])
	r1_seq[20] = reverse_complement(r1_seq[20])
	read1.sequence = "".join(r1_seq)
	read1_merged, read2_merged = trimmer(read1, read2)
	assert read1_merged.merged
	assert read2_merged is None
	assert len(read1_merged) == 26
	assert read1_merged.sequence == 'CCAAGCAGACTTTCACTCAGTTTGCA'
	assert read1_merged.qualities == ('#' * 24) + ( '!' * 2)
	
	# too few overlapping bases
	read1.merged = False
	r1_seq[15] = reverse_complement(r1_seq[15])
	read1.sequence = "".join(r1_seq)
	read1_merged, read2_merged = trimmer(read1, read2)
	assert read1_merged.merged is False
	assert read2 is not None



# coding: utf-8
from __future__ import print_function, division, absolute_import

from atropos.colorspace import encode, decode
from atropos.scripts.atropos import main
from .utils import run, datapath

# If there are any unknown characters in the test sequence,
# round tripping will only work if all characters after the
# first unknown character are also unknown:
# encode("TNGN") == "T444", but
# decode("T444") == "TNNN".

sequences = [
	"",
	"C",
	"ACGGTC",
	"TN",
	"TN.",
	"TNN.N",
	"CCGGCAGCATTCATTACGACAACGTGGCACCGTGTTTTCTCGGTGGTA",
	"TGCAGTTGATGATCGAAGAAAACGACATCATCAGCCAGCAAGTGC",
	"CAGGGTTTGATGAGTGGCTGTGGGTGCTGGCGTATCCGGG"
	]


def test_encode():
	assert encode("AA") == "A0"
	assert encode("AC") == "A1"
	assert encode("AG") == "A2"
	assert encode("AT") == "A3"
	assert encode("CA") == "C1"
	assert encode("CC") == "C0"
	assert encode("CG") == "C3"
	assert encode("CT") == "C2"
	assert encode("GA") == "G2"
	assert encode("GC") == "G3"
	assert encode("GG") == "G0"
	assert encode("GT") == "G1"
	assert encode("TA") == "T3"
	assert encode("TC") == "T2"
	assert encode("TG") == "T1"
	assert encode("TT") == "T0"

	assert encode("TN") == "T4"
	assert encode("NT") == "N4"
	assert encode("NN") == "N4"

	assert encode("ACGGTC") == "A13012"
	assert encode("TTT.N") == "T0044"
	assert encode("TTNT.N") == "T04444"


def test_decode():
	for s in sequences:
		expected = s.replace('.', 'N')
		encoded = encode(s)
		assert decode(encoded) == expected
	assert decode('A.') == 'AN'
	assert decode('C.') == 'CN'
	assert decode('G.') == 'GN'
	assert decode('T.') == 'TN'


def test_qualtrim_csfastaqual():
	'''-q with csfasta/qual files'''
	run("-c -q 10", "solidqual.fastq", "solid.csfasta", 'solid.qual')


def test_E3M():
	'''Read the E3M dataset'''
	# not really colorspace, but a fasta/qual file pair
	main(['-o', '/dev/null', datapath("E3M.fasta"), datapath("E3M.qual")])


def test_bwa():
	'''MAQ-/BWA-compatible output'''
	run("-c -e 0.12 -a 330201030313112312 -x 552: --maq", "solidmaq.fastq", "solid.csfasta", 'solid.qual')


def test_bfast():
	'''BFAST-compatible output'''
	run("-c -e 0.12 -a 330201030313112312 -x abc: --strip-f3", "solidbfast.fastq", "solid.csfasta", 'solid.qual')


def test_trim_095():
	'''some reads properly trimmed since atropos 0.9.5'''
	run("-c -e 0.122 -a 330201030313112312", "solid.fasta", "solid.fasta")


def test_solid():
	run("-c -e 0.122 -a 330201030313112312", "solid.fastq", "solid.fastq")


def test_solid_basespace_adapter():
	'''colorspace adapter given in basespace'''
	run("-c -e 0.122 -a CGCCTTGGCCGTACAGCAG", "solid.fastq", "solid.fastq")


def test_solid5p():
	'''test 5' colorspace adapter'''
	# this is not a real adapter, just a random string
	# in colorspace: C0302201212322332333
	run("-c -e 0.1 --trim-primer -g CCGGAGGTCAGCTCGCTATA", "solid5p.fasta", "solid5p.fasta")


def test_solid5p_prefix_notrim():
	'''test anchored 5' colorspace adapter, no primer trimming'''
	run("-c -e 0.1 -g ^CCGGAGGTCAGCTCGCTATA", "solid5p-anchored.notrim.fasta", "solid5p.fasta")


def test_solid5p_prefix():
	'''test anchored 5' colorspace adapter'''
	run("-c -e 0.1 --trim-primer -g ^CCGGAGGTCAGCTCGCTATA", "solid5p-anchored.fasta", "solid5p.fasta")


def test_solid5p_fastq():
	'''test 5' colorspace adapter'''
	# this is not a real adapter, just a random string
	# in colorspace: C0302201212322332333
	run("-c -e 0.1 --trim-primer -g CCGGAGGTCAGCTCGCTATA", "solid5p.fastq", "solid5p.fastq")


def test_solid5p_prefix_notrim_fastq():
	'''test anchored 5' colorspace adapter, no primer trimming'''
	run("-c -e 0.1 -g ^CCGGAGGTCAGCTCGCTATA", "solid5p-anchored.notrim.fastq", "solid5p.fastq")


def test_solid5p_prefix_fastq():
	'''test anchored 5' colorspace adapter'''
	run("-c -e 0.1 --trim-primer -g ^CCGGAGGTCAGCTCGCTATA", "solid5p-anchored.fastq", "solid5p.fastq")


def test_sra_fastq():
	'''test SRA-formatted colorspace FASTQ'''
	run("-c -e 0.1 --format sra-fastq -a CGCCTTGGCCGTACAGCAG", "sra.fastq", "sra.fastq")


def test_no_zero_cap():
	run("--no-zero-cap -c -e 0.122 -a CGCCTTGGCCGTACAGCAG", "solid-no-zerocap.fastq", "solid.fastq")



# coding: utf-8
from __future__ import print_function, division, absolute_import

import sys, os
from contextlib import contextmanager
from atropos.scripts import atropos

@contextmanager
def redirect_stderr():
	"Send stderr to stdout. Nose doesn't capture stderr, yet."
	old_stderr = sys.stderr
	sys.stderr = sys.stdout
	yield
	sys.stderr = old_stderr


@contextmanager
def temporary_path(name):
	directory = os.path.join(os.path.dirname(__file__), 'testtmp')
	if not os.path.isdir(directory):
		os.mkdir(directory)
	path = os.path.join(directory, name)
	yield path
	os.remove(path)


def datapath(path):
	return os.path.join(os.path.dirname(__file__), 'data', path)


def cutpath(path):
	return os.path.join(os.path.dirname(__file__), 'cut', path)


def files_equal(path1, path2):
	#return os.system("diff -u {0} {1}".format(path1, path2)) == 0
	with open(path1,'r') as i1, open(path2, 'r') as i2:
		print("<[{}]>".format(i1.read()))
		print("<[{}]>".format(i2.read()))
	from subprocess import check_output, CalledProcessError
	try:
		check_output("diff -u {0} {1}".format(path1, path2), shell=True)
		return True
	except CalledProcessError as e:
		print("Diff: <{}>".format(e.output.decode("utf-8")))
		return False


def run(params, expected, inpath, inpath2=None):
	if type(params) is str:
		params = params.split()
	with temporary_path(expected) as tmp_fastaq:
		params += ['-o', tmp_fastaq ] # TODO not parallelizable
		params += [ datapath(inpath) ]
		if inpath2:
			params += [ datapath(inpath2) ]
		print(params)
		assert atropos.main(params) is None
		# TODO redirect standard output
		assert files_equal(cutpath(expected), tmp_fastaq)
	# TODO diff log files



# kate: syntax Python;
# cython: profile=False, emit_code_comments=False
"""
Quality trimming.
"""

def quality_trim_index(str qualities, int cutoff_front, int cutoff_back, int base=33):
    """
    Find the positions at which to trim low-quality ends from a nucleotide sequence.
    Return tuple (start, stop) that indicates the good-quality segment.

    Qualities are assumed to be ASCII-encoded as chr(qual + base).

    The algorithm is the same as the one used by BWA within the function
    'bwa_trim_read':
    - Subtract the cutoff value from all qualities.
    - Compute partial sums from all indices to the end of the sequence.
    - Trim sequence at the index at which the sum is minimal.
    """
    cdef int s
    cdef int max_qual
    cdef int stop = len(qualities)
    cdef int start = 0
    cdef int i

    # find trim position for 5' end
    s = 0
    max_qual = 0
    for i in range(len(qualities)):
        s += cutoff_front - (ord(qualities[i]) - base)
        if s < 0:
            break
        if s > max_qual:
            max_qual = s
            start = i + 1

    # same for 3' end
    max_qual = 0
    s = 0
    for i in reversed(xrange(len(qualities))):
        s += cutoff_back - (ord(qualities[i]) - base)
        if s < 0:
            break
        if s > max_qual:
            max_qual = s
            stop = i
    if start >= stop:
        start, stop = 0, 0
    return (start, stop)


def nextseq_trim_index(sequence, int cutoff, int base=33):
    """
    Variant of the above quality trimming routine that works on NextSeq data.
    With Illumina NextSeq, bases are encoded with two colors. 'No color' (a
    dark cycle) usually means that a 'G' was sequenced, but that also occurs
    when sequencing falls off the end of the fragment. The read then contains
    a run of high-quality G bases in the end.

    This routine works as the one above, but counts qualities belonging to 'G'
    bases as being equal to cutoff - 1.
    """
    bases = sequence.sequence
    qualities = sequence.qualities
    cdef:
        int s = 0
        int max_qual = 0
        int max_i = len(qualities)
        int i, q

    s = 0
    max_qual = 0
    max_i = len(qualities)
    for i in reversed(xrange(max_i)):
        q = ord(qualities[i]) - base
        if bases[i] == 'G':
            q = cutoff - 1
        s += cutoff - q
        if s < 0:
            break
        if s > max_qual:
            max_qual = s
            max_i = i
    return max_i



# coding: utf-8
"""
Colorspace conversion routines.

Inspired by agapython/util/Dibase.py from Corona lite,
but reimplemented to avoid licensing issues.

Encoding Table

  A C G T
A 0 1 2 3
C 1 0 3 2
G 2 3 0 1
T 3 2 1 0
"""
def _initialize_dicts():
    """
    Create the colorspace encoding and decoding dictionaries.
    """
    enc = {}
    for i, c1 in enumerate("ACGT"):
        enc['N' + c1] = '4'
        enc[c1 + 'N'] = '4'
        enc['.' + c1] = '4'
        enc[c1 + '.'] = '4'
        for j, c2 in enumerate("ACGT"):
            # XOR of nucleotides gives color
            enc[c1 + c2] = chr(ord('0') + (i ^ j))
    enc.update({ 'NN': '4', 'N.': '4', '.N': '4', '..': '4'})

    dec = {}
    for i, c1 in enumerate("ACGT"):
        dec['.' + str(i)] = 'N'
        dec['N' + str(i)] = 'N'
        dec[c1 + '4'] = 'N'
        dec[c1 + '.'] = 'N'
        for j, c2 in enumerate("ACGT"):
            # XOR of nucleotides gives color
            dec[c1 + chr(ord('0') + (i ^ j))] = c2
    dec['N4'] = 'N'

    return (enc, dec)

def encode(s):
    """
    Given a sequence of nucleotides, convert them to
    colorspace. Only uppercase characters are allowed.
    >>> encode("ACGGTC")
    "A13012"
    """
    if not s:
        return s
    r = s[0:1]
    for i in range(len(s) - 1):
        r += ENCODE[s[i:i+2]]
    return r

def decode(s):
    """
    Decode a sequence of colors to nucleotide space.
    The first character in s must be a nucleotide.
    Only uppercase characters are allowed.
    >>> decode("A13012")
    "ACGGTC"
    """
    if len(s) < 2:
        return s
    x = s[0]
    result = x
    for c in s[1:]:
        x = DECODE[x + c]
        result += x
    return result

(ENCODE, DECODE) = _initialize_dicts()



"""
Open compressed files transparently.
"""
import gzip
import sys
import io
import os
import re
from subprocess import Popen, PIPE

try:
    import bz2
except ImportError:
    bz2 = None

try:
    import lzma
except ImportError:
    lzma = None

def open_output(filename, mode='w'):
    """
    Replacement for the "open" function that can also open files that have
    been compressed with gzip, bzip2 or xz. If the filename is '-', standard
    output (mode 'w') or input (mode 'r') is returned. If the filename ends
    with .gz, the file is opened with a pipe to the gzip program. If that
    does not work, then gzip.open() is used (the gzip module is slower than
    the pipe to the gzip program). If the filename ends with .bz2, it's
    opened as a bz2.BZ2File. Otherwise, the regular open() is used.

    mode can be: 'a', 'ab', 'wt', or 'wb'
    Instead of 'rt' and 'wt', 'r' and 'w' can be used as abbreviations.

    In Python 2, the 't' and 'b' characters are ignored.

    Append mode ('a') is unavailable with BZ2 compression and will raise an error.
    """
    if mode == 'w':
        mode = 'wt'
    elif mode == 'a':
        mode = 'at'
    if mode not in ('wt', 'wb', 'at', 'ab'):
        raise ValueError("mode '{0}' not supported".format(mode))
    if not isinstance(filename, str):
        raise ValueError("the filename must be a string")

    # standard input and standard output handling
    if filename == '-':
        return dict(
            wt=sys.stdout,
            wb=sys.stdout.buffer)[mode]
    
    return open(filename, mode)

def xopen(filename, mode='r'):
    """
    Replacement for the "open" function that can also open files that have
    been compressed with gzip, bzip2 or xz. If the filename is '-', standard
    output (mode 'w') or input (mode 'r') is returned. If the filename ends
    with .gz, the file is opened with a pipe to the gzip program. If that
    does not work, then gzip.open() is used (the gzip module is slower than
    the pipe to the gzip program). If the filename ends with .bz2, it's
    opened as a bz2.BZ2File. Otherwise, the regular open() is used.

    mode can be: 'rt', 'rb', 'a', 'wt', or 'wb'
    Instead of 'rt' and 'wt', 'r' and 'w' can be used as abbreviations.

    In Python 2, the 't' and 'b' characters are ignored.

    Append mode ('a') is unavailable with BZ2 compression and will raise an error.
    """
    if mode == 'r':
        mode = 'rt'
    elif mode == 'w':
        mode = 'wt'
    elif mode == 'a':
        mode = 'at'
    if mode not in ('rt', 'rb', 'wt', 'wb', 'at', 'ab'):
        raise ValueError("mode '{0}' not supported".format(mode))
    if not isinstance(filename, str):
        raise ValueError("the filename must be a string")

    # standard input and standard output handling
    if filename == '-':
        return dict(
            rt=sys.stdin,
            wt=sys.stdout,
            rb=sys.stdin.buffer,
            wb=sys.stdout.buffer)[mode]

    if filename.endswith('.bz2'):
        if bz2 is None:
            raise ImportError("Cannot open bz2 files: The bz2 module is not available")
        if 't' in mode:
            return io.TextIOWrapper(bz2.BZ2File(filename, mode[0]))
        else:
            return bz2.BZ2File(filename, mode)
    elif filename.endswith('.xz'):
        if lzma is None:
            raise ImportError("Cannot open xz files: The lzma module is not available "
                "(use Python 3.3 or newer)")
        return lzma.open(filename, mode)
    elif filename.endswith('.gz'):
        if 't' in mode:
            # gzip.open in Python 3.2 does not support modes 'rt' and 'wt''
            return io.TextIOWrapper(gzip.open(filename, mode[0]))
        else:
            if 'r' in mode:
                return io.BufferedReader(gzip.open(filename, mode))
            else:
                return io.BufferedWriter(gzip.open(filename, mode))
    else:
        return open(filename, mode)

compressors = {
    ".gz"  : gzip,
    ".bz2" : bz2,
    ".xz"  : lzma
}

def get_compressor(filename):
    for ext, compressor in compressors.items():
        if filename.endswith(ext):
            return compressor
    return None



# coding: utf-8
from __future__ import print_function, division, absolute_import

import shutil
from nose.tools import raises
from atropos.scripts import atropos
from .utils import run, files_equal, datapath, cutpath, redirect_stderr, temporary_path

def run_paired(params, in1, in2, expected1, expected2):
	if type(params) is str:
		params = params.split()
	with temporary_path('tmp1-' + expected1) as p1:
		with temporary_path('tmp2-' + expected2) as p2:
			params += ['-o', p1, '-p', p2]
			params += [datapath(in1), datapath(in2)]
			assert atropos.main(params) is None
			assert files_equal(cutpath(expected1), p1)
			assert files_equal(cutpath(expected2), p2)

def run_interleaved(params, inpath, expected):
	if type(params) is str:
		params = params.split()
	with temporary_path(expected) as tmp:
		params += ['--interleaved', '-o', tmp, datapath(inpath)]
		assert atropos.main(params) is None
		assert files_equal(cutpath(expected), tmp)


def run_interleaved2(params, inpath, expected1, expected2):
	assert False  # unused function
	if type(params) is str:
		params = params.split()
	with temporary_path('tmp1-' + expected1) as p1:
		with temporary_path('tmp2-' + expected2) as p2:
			params += ['--interleaved', '-o', p1, '-p', p2]
		params += [datapath(inpath)]
		assert atropos.main(params) is None
		assert files_equal(cutpath(expected), p1)
		assert files_equal(cutpath(expected), p2)


def test_paired_separate():
	'''test separate trimming of paired-end reads'''
	run('-a TTAGACATAT', 'paired-separate.1.fastq', 'paired.1.fastq')
	run('-a CAGTGGAGTA', 'paired-separate.2.fastq', 'paired.2.fastq')


def test_paired_end_legacy():
	'''--paired-output, not using -A/-B/-G'''
	# the -m 14 filters out one read, which should then also be filtered out in the second output file
	# -q 10 should not change anything: qualities in file 1 are high enough,
	# qualities in file 2 should not be inspected.
	run_paired('-a TTAGACATAT -m 14 -q 10',
		in1='paired.1.fastq', in2='paired.2.fastq',
		expected1='paired.m14.1.fastq', expected2='paired.m14.2.fastq'
	)


def test_untrimmed_paired_output():
	with temporary_path("tmp-untrimmed.1.fastq") as untrimmed1:
		with temporary_path("tmp-untrimmed.2.fastq") as untrimmed2:
			run_paired(['-a', 'TTAGACATAT',
				'--untrimmed-output', untrimmed1,
				'--untrimmed-paired-output', untrimmed2],
				in1='paired.1.fastq', in2='paired.2.fastq',
				expected1='paired-trimmed.1.fastq', expected2='paired-trimmed.2.fastq'
			)
			assert files_equal(cutpath('paired-untrimmed.1.fastq'), untrimmed1)
			assert files_equal(cutpath('paired-untrimmed.2.fastq'), untrimmed2)


def test_explicit_format_with_paired():
	# Use --format=fastq with input files whose extension is .txt
	with temporary_path("paired.1.txt") as txt1:
		with temporary_path("paired.2.txt") as txt2:
			shutil.copyfile(datapath("paired.1.fastq"), txt1)
			shutil.copyfile(datapath("paired.2.fastq"), txt2)
			run_paired('--format=fastq -a TTAGACATAT -m 14',
				in1=txt1, in2=txt2,
				expected1='paired.m14.1.fastq',
				expected2='paired.m14.2.fastq'
			)


def test_no_trimming_legacy():
	# make sure that this doesn't divide by zero
	atropos.main(['-a', 'XXXXX', '-o', '/dev/null', '-p', '/dev/null', datapath('paired.1.fastq'), datapath('paired.2.fastq')])


def test_no_trimming():
	# make sure that this doesn't divide by zero
	atropos.main(['-a', 'XXXXX', '-A', 'XXXXX', '-o', '/dev/null', '-p', '/dev/null', datapath('paired.1.fastq'), datapath('paired.2.fastq')])


@raises(SystemExit)
def test_missing_file():
	with redirect_stderr():
		atropos.main(['-a', 'XX', '--paired-output', 'out.fastq', datapath('paired.1.fastq')])


@raises(SystemExit)
def test_first_too_short():
	with temporary_path("truncated.1.fastq") as trunc1:
		# Create a truncated file in which the last read is missing
		with open(datapath('paired.1.fastq')) as f:
			lines = f.readlines()
			lines = lines[:-4]
		with open(trunc1, 'w') as f:
			f.writelines(lines)
		with redirect_stderr():
			atropos.main('-a XX --paired-output out.fastq'.split() + [trunc1, datapath('paired.2.fastq')])


@raises(SystemExit)
def test_second_too_short():
	with temporary_path("truncated.2.fastq") as trunc2:
		# Create a truncated file in which the last read is missing
		with open(datapath('paired.2.fastq')) as f:
			lines = f.readlines()
			lines = lines[:-4]
		with open(trunc2, 'w') as f:
			f.writelines(lines)
		with redirect_stderr():
			atropos.main('-a XX --paired-output out.fastq'.split() + [datapath('paired.1.fastq'), trunc2])


@raises(SystemExit)
def test_unmatched_read_names():
	with temporary_path("swapped.1.fastq") as swapped:
		# Create a file in which reads 2 and are swapped
		with open(datapath('paired.1.fastq')) as f:
			lines = f.readlines()
			lines = lines[0:4] + lines[8:12] + lines[4:8] + lines[12:]
		with open(swapped, 'w') as f:
			f.writelines(lines)
		with redirect_stderr():
			atropos.main('-a XX -o out1.fastq --paired-output out2.fastq'.split() + [swapped, datapath('paired.2.fastq')])


def test_legacy_minlength():
	'''Ensure -m is not applied to second read in a pair in legacy mode'''
	run_paired('-a XXX -m 27',
		in1='paired.1.fastq', in2='paired.2.fastq',
		expected1='paired-m27.1.fastq', expected2='paired-m27.2.fastq'
	)


def test_paired_end():
	'''single-pass paired-end with -m'''
	run_paired('-a TTAGACATAT -A CAGTGGAGTA -m 14',
		in1='paired.1.fastq', in2='paired.2.fastq',
		expected1='paired.1.fastq', expected2='paired.2.fastq'
	)


def test_paired_anchored_back_no_indels():
	run_paired("-a BACKADAPTER$ -A BACKADAPTER$ -N --no-indels",
		in1='anchored-back.fasta', in2='anchored-back.fasta',
		expected1='anchored-back.fasta', expected2="anchored-back.fasta"
	)


def test_paired_end_qualtrim():
	'''single-pass paired-end with -q and -m'''
	run_paired('-q 20 -a TTAGACATAT -A CAGTGGAGTA -m 14 -M 90',
		in1='paired.1.fastq', in2='paired.2.fastq',
		expected1='pairedq.1.fastq', expected2='pairedq.2.fastq'
	)


def test_paired_end_qualtrim_swapped():
	'''single-pass paired-end with -q and -m, but files swapped'''
	run_paired('-q 20 -a CAGTGGAGTA -A TTAGACATAT -m 14',
		in1='paired.2.fastq', in2='paired.1.fastq',
		expected1='pairedq.2.fastq', expected2='pairedq.1.fastq'
	)


def test_paired_end_cut():
	run_paired('-u 3 -u -1 -U 4 -U -2',
		in1='paired.1.fastq', in2='paired.2.fastq',
		expected1='pairedu.1.fastq', expected2='pairedu.2.fastq'
	)


def test_paired_end_A_only():
	run_paired('-A CAGTGGAGTA',
		in1='paired.1.fastq', in2='paired.2.fastq',
		expected1='paired-onlyA.1.fastq', expected2='paired-onlyA.2.fastq'
	)


def test_discard_untrimmed():
	# issue #146
	# the first adapter is a sequence cut out from the first read
	run_paired('-a CTCCAGCTTAGACATATC -A XXXXXXXX --discard-untrimmed',
		in1='paired.1.fastq', in2='paired.2.fastq',
		expected1='empty.fastq', expected2='empty.fastq'
	)


def test_discard_trimmed():
	run_paired('-A C -O 1 --discard-trimmed',  # applies everywhere
		in1='paired.1.fastq', in2='paired.2.fastq',
		expected1='empty.fastq', expected2='empty.fastq'
	)


def test_interleaved():
	'''single-pass interleaved paired-end with -q and -m'''
	run_interleaved('-q 20 -a TTAGACATAT -A CAGTGGAGTA -m 14 -M 90',
		inpath='interleaved.fastq', expected='interleaved.fastq'
	)


@raises(SystemExit)
def test_interleaved_no_paired_output():
	with temporary_path("temp-paired.1.fastq") as p1:
		with temporary_path("temp-paired.2.fastq") as p2:
			params = '-a XX --interleaved'.split()
			with redirect_stderr():
				params += [ '-o', p1, '-p1', p2, 'paired.1.fastq', 'paired.2.fastq']
				atropos.main(params)

"""
def test_interleaved_input_paired_output():
	'''single-pass interleaved paired-end with -q and -m, paired output'''
	run_interleaved2('-q 20 -a TTAGACATAT -A CAGTGGAGTA -m 14 -M 90',
		inpath='interleaved.fastq', expected1='pairedq1.fastq', expected2='pairedq2.fastq'
	)
"""


def test_pair_filter():
	run_paired('--pair-filter=both -a TTAGACATAT -A GGAGTA -m 14',
		in1='paired.1.fastq', in2='paired.2.fastq',
		expected1='paired-filterboth.1.fastq', expected2='paired-filterboth.2.fastq'
	)


def test_too_short_paired_output():
	with temporary_path("temp-too-short.1.fastq") as p1:
		with temporary_path("temp-too-short.2.fastq") as p2:
			run_paired('-a TTAGACATAT -A CAGTGGAGTA -m 14 --too-short-output '
				'{0} --too-short-paired-output {1}'.format(p1, p2),
				in1='paired.1.fastq', in2='paired.2.fastq',
				expected1='paired.1.fastq', expected2='paired.2.fastq'
			)
			assert files_equal(cutpath('paired-too-short.1.fastq'), p1)
			assert files_equal(cutpath('paired-too-short.2.fastq'), p2)


def test_too_long_output():
	with temporary_path("temp-too-long.1.fastq") as p1:
		with temporary_path("temp-too-long.2.fastq") as p2:
			run_paired('-a TTAGACATAT -A CAGTGGAGTA -M 14 --too-long-output '
				'{0} --too-long-paired-output {1}'.format(p1, p2),
				in1='paired.1.fastq', in2='paired.2.fastq',
				expected1='paired-too-short.1.fastq', expected2='paired-too-short.2.fastq'
			)
			assert files_equal(cutpath('paired.1.fastq'), p1)
			assert files_equal(cutpath('paired.2.fastq'), p2)


@raises(SystemExit)
def test_too_short_output_paired_option_missing():
	with temporary_path("temp-too-short.1.fastq") as p1:
		run_paired('-a TTAGACATAT -A CAGTGGAGTA -m 14 --too-short-output '
			'{0}'.format(p1),
			in1='paired.1.fastq', in2='paired.2.fastq',
			expected1='paired.1.fastq', expected2='paired.2.fastq'
		)

def test_custom_bisulfite_1():
	run_paired('-a TTAGACATAT -A CAGTGGAGTA -m 14 -q 0 --bisulfite 2,2,1,1',
		in1='paired.1.fastq', in2='paired.2.fastq',
		expected1='paired.1.fastq', expected2='paired.2.fastq'
	)

def test_custom_bisulfite_2():
	run_paired('-a TTAGACATAT -A CAGTGGAGTA -m 10 -q 0 --bisulfite 20,20,1,1;0,0,0,0',
		in1='paired.1.fastq', in2='paired.2.fastq',
		expected1='paired_bis1_1.fastq', expected2='paired_bis1_2.fastq'
	)



# coding: utf-8
from __future__ import print_function, division, absolute_import
import sys

__version__ = '1.0.2'

def check_importability():  # pragma: no cover
    try:
        import atropos._align
    except ImportError as e:
        if 'undefined symbol' in str(e):
            print("""
ERROR: A required extension module could not be imported because it is
incompatible with your system. A quick fix is to recompile the extension
modules with the following command:

    {0} setup.py build_ext -i

See the documentation for alternative ways of installing the program.

The original error message follows.
""".format(sys.executable))
        raise



# coding: utf-8
"""
Adapters
"""
import sys
import re
from collections import defaultdict, namedtuple
from atropos import align, colorspace
from atropos.seqio import ColorspaceSequence, FastaReader

# Constants for the find_best_alignment function.
# The function is called with SEQ1 as the adapter, SEQ2 as the read.
# TODO get rid of those constants, use strings instead
BACK = align.START_WITHIN_SEQ2 | align.STOP_WITHIN_SEQ2 | align.STOP_WITHIN_SEQ1
FRONT = align.START_WITHIN_SEQ2 | align.STOP_WITHIN_SEQ2 | align.START_WITHIN_SEQ1
PREFIX = align.STOP_WITHIN_SEQ2
SUFFIX = align.START_WITHIN_SEQ2
ANYWHERE = align.SEMIGLOBAL
LINKED = 'linked'

DEFAULT_ADAPTERS = dict(
    nextera='CTGTCTCTTATA',
    illumina='AGATCGGAAGAGC'
)

# TODO: write test
def get_sequence_for_spec(spec):
    return DEFAULT_ADAPTERS.get(spec.lower(), spec)

def parse_braces(sequence):
    """
    Replace all occurrences of ``x{n}`` (where x is any character) with n
    occurrences of x. Raise ValueError if the expression cannot be parsed.

    >>> parse_braces('TGA{5}CT')
    TGAAAAACT
    """
    # Simple DFA with four states, encoded in prev
    result = ''
    prev = None
    for s in re.split('(\{|\})', sequence):
        if s == '':
            continue
        if prev is None:
            if s == '{':
                raise ValueError('"{" must be used after a character')
            if s == '}':
                raise ValueError('"}" cannot be used here')
            prev = s
            result += s
        elif prev == '{':
            prev = int(s)
            if not 0 <= prev <= 10000:
                raise ValueError('Value {} invalid'.format(prev))
        elif isinstance(prev, int):
            if s != '}':
                raise ValueError('"}" expected')
            result = result[:-1] + result[-1] * prev
            prev = None
        else:
            if s != '{':
                raise ValueError('Expected "{"')
            prev = '{'
    # Check if we are in a non-terminating state
    if isinstance(prev, int) or prev == '{':
        raise ValueError("Unterminated expression")
    return result

class AdapterParser(object):
    """
    Factory for Adapter classes that all use the same parameters (error rate,
    indels etc.). The given **kwargs will be passed to the Adapter constructors.
    """
    def __init__(self, colorspace=False, **kwargs):
        self.colorspace = colorspace
        self.constructor_args = kwargs
        self.adapter_class = ColorspaceAdapter if colorspace else Adapter

    def parse(self, spec, name=None, cmdline_type='back'):
        """
        Parse an adapter specification not using ``file:`` notation and return
        an object of an appropriate Adapter class. The notation for anchored
        5' and 3' adapters is supported. If the name parameter is None, then
        an attempt is made to extract the name from the specification
        (If spec is 'name=ADAPTER', name will be 'name'.)

        cmdline_type -- describes which commandline parameter was used (``-a``
        is 'back', ``-b`` is 'anywhere', and ``-g`` is 'front').
        """
        if name is None:
            name, spec = self._extract_name(spec)
        sequence = get_sequence_for_spec(spec)
        types = dict(back=BACK, front=FRONT, anywhere=ANYWHERE)
        if cmdline_type not in types:
            raise ValueError('cmdline_type cannot be {0!r}'.format(cmdline_type))
        where = types[cmdline_type]
        if where == FRONT and spec.startswith('^'):  # -g ^ADAPTER
            sequence, where = spec[1:], PREFIX
        elif where == BACK:
            sequence1, middle, sequence2 = spec.partition('...')
            if middle == '...':
                if not sequence1:  # -a ...ADAPTER
                    sequence = sequence1[3:]
                elif not sequence2:  # -a ADAPTER...
                    sequence, where = spec[:-3], PREFIX
                else:  # -a ADAPTER1...ADAPTER2
                    if self.colorspace:
                        raise NotImplementedError('Using linked adapters in colorspace is not supported')
                    if sequence1.startswith('^') or sequence2.endswith('$'):
                        raise NotImplementedError('Using "$" or "^" when '
                            'specifying a linked adapter is not supported')
                    return LinkedAdapter(sequence1, sequence2, name=name,
                        **self.constructor_args)
            elif spec.endswith('$'):   # -a ADAPTER$
                sequence, where = spec[:-1], SUFFIX
        if not sequence:
            raise ValueError("The adapter sequence is empty.")

        return self.adapter_class(sequence, where, name=name, **self.constructor_args)

    def parse_with_file(self, spec, cmdline_type='back'):
        """
        Parse an adapter specification and yield appropriate Adapter classes.
        This works like the parse() function above, but also supports the
        ``file:`` notation for reading adapters from an external FASTA
        file. Since a file can contain multiple adapters, this
        function is a generator.
        """
        if spec.startswith('file:'):
            # read adapter sequences from a file
            with FastaReader(spec[5:]) as fasta:
                for record in fasta:
                    name = record.name.split(None, 1)[0]
                    yield self.parse(record.sequence, name, cmdline_type)
        else:
            name, spec = self._extract_name(spec)
            yield self.parse(spec, name, cmdline_type)

    def _extract_name(self, spec):
        """
        Parse an adapter specification given as 'name=adapt' into 'name' and 'adapt'.
        """
        fields = spec.split('=', 1)
        if len(fields) > 1:
            name, spec = fields
            name = name.strip()
        else:
            name = None
        spec = spec.strip()
        return name, spec

    def parse_multi(self, back, anywhere, front):
        """
        Parse all three types of commandline options that can be used to
        specify adapters. back, anywhere and front are lists of strings,
        corresponding to the respective commandline types (-a, -b, -g).

        Return a list of appropriate Adapter classes.
        """
        adapters = []
        for specs, cmdline_type in (back, 'back'), (anywhere, 'anywhere'), (front, 'front'):
            for spec in specs:
                adapters.extend(self.parse_with_file(spec, cmdline_type))
        return adapters

MatchInfo = namedtuple("MatchInfo", (
    "read_name", "errors", "rstart", "rstop", "seq_before", "seq_adapter", "seq_after",
    "adapter_name", "qual_before", "qual_adapter", "qual_after", "is_front", "asize",
    "rsize_adapter", "rsize_total"
))

class Match(object):
    """
    TODO creating instances of this class is relatively slow and responsible for quite some runtime.
    """
    __slots__ = ['astart', 'astop', 'rstart', 'rstop', 'matches', 'errors', 'front', 'adapter', 'read', 'length']
    def __init__(self, astart, astop, rstart, rstop, matches, errors, front, adapter, read):
        self.astart = astart
        self.astop = astop
        self.rstart = rstart
        self.rstop = rstop
        self.matches = matches
        self.errors = errors
        self.front = self._guess_is_front() if front is None else front
        self.adapter = adapter
        self.read = read
        # Number of aligned characters in the adapter. If there are
        # indels, this may be different from the number of characters
        # in the read.
        self.length = self.astop - self.astart
        assert self.length > 0
        assert self.errors / self.length <= self.adapter.max_error_rate
        assert self.length - self.errors > 0

    def __str__(self):
        return 'Match(astart={0}, astop={1}, rstart={2}, rstop={3}, matches={4}, errors={5})'.format(
            self.astart, self.astop, self.rstart, self.rstop, self.matches, self.errors)

    def _guess_is_front(self):
        """
        Return whether this is guessed to be a front adapter.

        The match is assumed to be a front adapter when the first base of
        the read is involved in the alignment to the adapter.
        """
        return self.rstart == 0

    def wildcards(self, wildcard_char='N'):
        """
        Return a string that contains, for each wildcard character,
        the character that it matches. For example, if the adapter
        ATNGNA matches ATCGTA, then the string 'CT' is returned.

        If there are indels, this is not reliable as the full alignment
        is not available.
        """
        wildcards = [ self.read.sequence[self.rstart + i:self.rstart + i + 1] for i in range(self.length)
            if self.adapter.sequence[self.astart + i] == wildcard_char and self.rstart + i < len(self.read.sequence) ]
        return ''.join(wildcards)

    def rest(self):
        """
        Return the part of the read before this match if this is a
        'front' (5') adapter,
        return the part after the match if this is not a 'front' adapter (3').
        This can be an empty string.
        """
        if self.front:
            return self.read.sequence[:self.rstart]
        else:
            return self.read.sequence[self.rstop:]
    
    # TODO: write test
    def get_info_record(self):
        seq = self.read.sequence
        qualities = self.read.qualities
        if qualities is None:
            qualities = ''
        rsize = rsize_total = self.rstop - self.rstart
        if self.front and self.rstart > 0:
            rsize_total = self.rstop
        elif not self.front and self.rstop < len(seq):
            rsize_total = len(seq) - self.rstart
        return MatchInfo(
            self.read.name,
            self.errors,
            self.rstart,
            self.rstop,
            seq[0:self.rstart],
            seq[self.rstart:self.rstop],
            seq[self.rstop:],
            self.adapter.name,
            qualities[0:self.rstart],
            qualities[self.rstart:self.rstop],
            qualities[self.rstop:],
            self.front,
            self.astop - self.astart,
            rsize, rsize_total
        )

def _generate_adapter_name(_start=[1]):
    name = str(_start[0])
    _start[0] += 1
    return name

class Adapter(object):
    """
    An adapter knows how to match itself to a read.
    In particular, it knows where it should be within the read and how to interpret
    wildcard characters.

    where --  One of the BACK, FRONT, PREFIX, SUFFIX or ANYWHERE constants.
        This influences where the adapter is allowed to appear within in the
        read and also which part of the read is removed.

    sequence -- The adapter sequence as string. Will be converted to uppercase.
        Also, Us will be converted to Ts.

    max_error_rate -- Maximum allowed error rate. The error rate is
        the number of errors in the alignment divided by the length
        of the part of the alignment that matches the adapter.

    minimum_overlap -- Minimum length of the part of the alignment
        that matches the adapter.

    read_wildcards -- Whether IUPAC wildcards in the read are allowed.

    adapter_wildcards -- Whether IUPAC wildcards in the adapter are
        allowed.

    name -- optional name of the adapter. If not provided, the name is set to a
        unique number.
    """
    def __init__(self, sequence, where, max_error_rate=0.1, min_overlap=3,
            read_wildcards=False, adapter_wildcards=True, name=None, indels=True):
        self.debug = False
        self.name = _generate_adapter_name() if name is None else name
        self.sequence = parse_braces(sequence.upper().replace('U', 'T'))
        assert len(self.sequence) > 0
        self.where = where
        self.max_error_rate = max_error_rate
        self.min_overlap = min(min_overlap, len(self.sequence))
        self.indels = indels
        self.adapter_wildcards = adapter_wildcards and not set(self.sequence) <= set('ACGT')
        self.read_wildcards = read_wildcards
        # redirect trimmed() to appropriate function depending on adapter type
        trimmers = {
            FRONT: self._trimmed_front,
            PREFIX: self._trimmed_front,
            BACK: self._trimmed_back,
            SUFFIX: self._trimmed_back,
            ANYWHERE: self._trimmed_anywhere
        }
        self.trimmed = trimmers[where]
        if where == ANYWHERE:
            self._front_flag = None  # means: guess
        else:
            self._front_flag = where not in (BACK, SUFFIX)
        # statistics about length of removed sequences
        self.lengths_front = defaultdict(int)
        self.lengths_back = defaultdict(int)
        self.errors_front = defaultdict(lambda: defaultdict(int))
        self.errors_back = defaultdict(lambda: defaultdict(int))
        self.adjacent_bases = { 'A': 0, 'C': 0, 'G': 0, 'T': 0, '': 0 }

        self.aligner = align.Aligner(self.sequence, self.max_error_rate,
            flags=self.where, wildcard_ref=self.adapter_wildcards, wildcard_query=self.read_wildcards)
        self.aligner.min_overlap = self.min_overlap
        if not self.indels:
            # TODO
            # When indels are disallowed, an entirely different algorithm
            # should be used.
            self.aligner.indel_cost = 100000

    def __repr__(self):
        return '<Adapter(name="{name}", sequence="{sequence}", where={where}, '\
            'max_error_rate={max_error_rate}, min_overlap={min_overlap}, '\
            'read_wildcards={read_wildcards}, '\
            'adapter_wildcards={adapter_wildcards}, '\
            'indels={indels})>'.format(**vars(self))

    def enable_debug(self):
        """
        Print out the dynamic programming matrix after matching a read to an
        adapter.
        """
        self.debug = True
        self.aligner.enable_debug()

    def match_to(self, read):
        """
        Attempt to match this adapter to the given read.

        Return an Match instance if a match was found;
        return None if no match was found given the matching criteria (minimum
        overlap length, maximum error rate).
        """
        read_seq = read.sequence.upper()
        pos = -1
        # try to find an exact match first unless wildcards are allowed
        if not self.adapter_wildcards:
            if self.where == PREFIX:
                pos = 0 if read_seq.startswith(self.sequence) else -1
            elif self.where == SUFFIX:
                pos = (len(read_seq) - len(self.sequence)) if read_seq.endswith(self.sequence) else -1
            else:
                pos = read_seq.find(self.sequence)
        if pos >= 0:
            match = Match(
                0, len(self.sequence), pos, pos + len(self.sequence),
                len(self.sequence), 0, self._front_flag, self, read)
        else:
            # try approximate matching
            if not self.indels and self.where in (PREFIX, SUFFIX):
                if self.where == PREFIX:
                    alignment = align.compare_prefixes(self.sequence, read_seq,
                        wildcard_ref=self.adapter_wildcards, wildcard_query=self.read_wildcards)
                else:
                    alignment = align.compare_suffixes(self.sequence, read_seq,
                        wildcard_ref=self.adapter_wildcards, wildcard_query=self.read_wildcards)
                astart, astop, rstart, rstop, matches, errors = alignment
                if astop - astart >= self.min_overlap and errors / (astop - astart) <= self.max_error_rate:
                    match = Match(*(alignment + (self._front_flag, self, read)))
                else:
                    match = None
            else:
                alignment = self.aligner.locate(read_seq)
                if self.debug:
                    print(self.aligner.dpmatrix)  # pragma: no cover
                if alignment is None:
                    match = None
                else:
                    astart, astop, rstart, rstop, matches, errors = alignment
                    match = Match(astart, astop, rstart, rstop, matches, errors, self._front_flag, self, read)

        if match is None:
            return None
        assert match.length > 0 and match.errors / match.length <= self.max_error_rate, match
        assert match.length >= self.min_overlap
        return match

    def _trimmed_anywhere(self, match):
        """Return a trimmed read"""
        if match.front:
            return self._trimmed_front(match)
        else:
            return self._trimmed_back(match)

    def _trimmed_front(self, match):
        """Return a trimmed read"""
        # TODO move away
        self.lengths_front[match.rstop] += 1
        self.errors_front[match.rstop][match.errors] += 1
        return match.read[match.rstop:]

    def _trimmed_back(self, match):
        """Return a trimmed read without the 3' (back) adapter"""
        # TODO move away
        self.lengths_back[len(match.read) - match.rstart] += 1
        self.errors_back[len(match.read) - match.rstart][match.errors] += 1
        adjacent_base = match.read.sequence[match.rstart-1:match.rstart]
        if adjacent_base not in 'ACGT':
            adjacent_base = ''
        self.adjacent_bases[adjacent_base] += 1
        return match.read[:match.rstart]

    def __len__(self):
        return len(self.sequence)


class ColorspaceAdapter(Adapter):
    def __init__(self, *args, **kwargs):
        super(ColorspaceAdapter, self).__init__(*args, **kwargs)
        has_nucleotide_seq = False
        if set(self.sequence) <= set('ACGT'):
            # adapter was given in basespace
            self.nucleotide_sequence = self.sequence
            has_nucleotide_seq = True
            self.sequence = colorspace.encode(self.sequence)[1:]
        if self.where in (PREFIX, FRONT) and not has_nucleotide_seq:
            raise ValueError("A 5' colorspace adapter needs to be given in nucleotide space")
        self.aligner.reference = self.sequence

    def match_to(self, read):
        """Return Match instance"""
        if self.where != PREFIX:
            return super(ColorspaceAdapter, self).match_to(read)
        # create artificial adapter that includes a first color that encodes the
        # transition from primer base into adapter
        asequence = colorspace.ENCODE[read.primer + self.nucleotide_sequence[0:1]] + self.sequence

        pos = 0 if read.sequence.startswith(asequence) else -1
        if pos >= 0:
            match = Match(
                0, len(asequence), pos, pos + len(asequence),
                len(asequence), 0, self._front_flag, self, read)
        else:
            # try approximate matching
            self.aligner.reference = asequence
            alignment = self.aligner.locate(read.sequence)
            if self.debug:
                print(self.aligner.dpmatrix)  # pragma: no cover
            if alignment is not None:
                match = Match(*(alignment + (self._front_flag, self, read)))
            else:
                match = None

        if match is None:
            return None
        assert match.length > 0 and match.errors / match.length <= self.max_error_rate
        assert match.length >= self.min_overlap
        return match

    def _trimmed_front(self, match):
        """Return a trimmed read"""
        read = match.read
        self.lengths_front[match.rstop] += 1
        self.errors_front[match.rstop][match.errors] += 1
        # to remove a front adapter, we need to re-encode the first color following the adapter match
        color_after_adapter = read.sequence[match.rstop:match.rstop + 1]
        if not color_after_adapter:
            # the read is empty
            return read[match.rstop:]
        base_after_adapter = colorspace.DECODE[self.nucleotide_sequence[-1:] + color_after_adapter]
        new_first_color = colorspace.ENCODE[read.primer + base_after_adapter]
        new_read = read[:]
        new_read.sequence = new_first_color + read.sequence[(match.rstop + 1):]
        new_read.qualities = read.qualities[match.rstop:] if read.qualities else None
        return new_read

    def _trimmed_back(self, match):
        """Return a trimmed read"""
        # trim one more color if long enough
        adjusted_rstart = max(match.rstart - 1, 0)
        self.lengths_back[len(match.read) - adjusted_rstart] += 1
        self.errors_back[len(match.read) - adjusted_rstart][match.errors] += 1
        return match.read[:adjusted_rstart]

    def __repr__(self):
        return '<ColorspaceAdapter(sequence={0!r}, where={1})>'.format(self.sequence, self.where)


class LinkedMatch(object):
    """
    Represent a match of a LinkedAdapter.

    TODO
    It shouldn’t be necessary to have both a Match and a LinkedMatch class.
    """
    def __init__(self, front_match, back_match, adapter):
        self.front_match = front_match
        self.back_match = back_match
        self.adapter = adapter
        assert front_match is not None
    
    def get_info_record(self):
        return self.back_match.get_info_record() if self.back_match else self.front_match.get_info_record()


class LinkedAdapter(object):
    """
    """
    def __init__(self, front_sequence, back_sequence,
            front_anchored=True, back_anchored=False, name=None, **kwargs):
        """
        kwargs are passed on to individual Adapter constructors
        """
        assert front_anchored and not back_anchored
        where1 = PREFIX if front_anchored else FRONT
        where2 = SUFFIX if back_anchored else BACK
        self.front_anchored = front_anchored
        self.back_anchored = back_anchored

        # The following attributes are needed for the report
        self.where = LINKED
        self.name = _generate_adapter_name() if name is None else name
        self.front_adapter = Adapter(front_sequence, where=where1, name=None, **kwargs)
        self.back_adapter = Adapter(back_sequence, where=where2, name=None, **kwargs)

    def enable_debug(self):
        self.front_adapter.enable_debug()
        self.back_adapter.enable_debug()

    def match_to(self, read):
        """
        Match the linked adapters against the given read. If the 'front' adapter
        is not found, the 'back' adapter is not searched for.
        """
        front_match = self.front_adapter.match_to(read)
        if front_match is None:
            return None
        # TODO use match.trimmed() instead as soon as that does not update
        # statistics anymore
        read = read[front_match.rstop:]
        back_match = self.back_adapter.match_to(read)
        return LinkedMatch(front_match, back_match, self)

    def trimmed(self, match):
        front_trimmed = self.front_adapter.trimmed(match.front_match)
        if match.back_match:
            return self.back_adapter.trimmed(match.back_match)
        else:
            return front_trimmed

    # Lots of forwarders (needed for the report). I’m sure this can be done
    # in a better way.

    @property
    def lengths_front(self):
        return self.front_adapter.lengths_front

    @property
    def lengths_back(self):
        return self.back_adapter.lengths_back

    @property
    def errors_front(self):
        return self.front_adapter.errors_front

    @property
    def errors_back(self):
        return self.back_adapter.errors_back

    @property
    def adjacent_bases(self):
        return self.back_adapter.adjacent_bases



# coding: utf-8
# TODO
# test with the --output option
# test reading from standard input
from __future__ import print_function, division, absolute_import

import os
import sys
from io import StringIO
from nose.tools import raises
from atropos.scripts import atropos
from .utils import run, files_equal, datapath, cutpath, redirect_stderr, temporary_path

def test_example():
	run('-N -b ADAPTER', 'example.fa', 'example.fa')

def test_small():
	run('-b TTAGACATATCTCCGTCG', 'small.fastq', 'small.fastq')

def test_empty():
	'''empty input'''
	run('-a TTAGACATATCTCCGTCG', 'empty.fastq', 'empty.fastq')

def test_newlines():
	'''DOS/Windows newlines'''
	run('-e 0.12 -b TTAGACATATCTCCGTCG', 'dos.fastq', 'dos.fastq')

def test_lowercase():
	'''lowercase adapter'''
	run('-b ttagacatatctccgtcg', 'lowercase.fastq', 'small.fastq')


def test_rest():
	'''-r/--rest-file'''
	with temporary_path('rest.tmp') as rest_tmp:
		run(['-b', 'ADAPTER', '-N', '-r', rest_tmp], "rest.fa", "rest.fa")
		assert files_equal(datapath('rest.txt'), rest_tmp)


def test_restfront():
	with temporary_path("rest.txt") as path:
		run(['-g', 'ADAPTER', '-N', '-r', path], "restfront.fa", "rest.fa")
		assert files_equal(datapath('restfront.txt'), path)


def test_discard():
	'''--discard'''
	run("-b TTAGACATATCTCCGTCG --discard", "discard.fastq", "small.fastq")


def test_discard_untrimmed():
	'''--discard-untrimmed'''
	run('-b CAAGAT --discard-untrimmed', 'discard-untrimmed.fastq', 'small.fastq')


def test_plus():
	'''test if sequence name after the "+" is retained'''
	run("-e 0.12 -b TTAGACATATCTCCGTCG", "plus.fastq", "plus.fastq")


def test_extensiontxtgz():
	'''automatic recognition of "_sequence.txt.gz" extension'''
	run("-b TTAGACATATCTCCGTCG", "s_1_sequence.txt", "s_1_sequence.txt.gz")


def test_format():
	'''the -f/--format parameter'''
	run("-f fastq -b TTAGACATATCTCCGTCG", "small.fastq", "small.myownextension")


def test_minimum_length():
	'''-m/--minimum-length'''
	run("-c -m 5 -a 330201030313112312", "minlen.fa", "lengths.fa")


def test_too_short():
	'''--too-short-output'''
	run("-c -m 5 -a 330201030313112312 --too-short-output tooshort.tmp.fa", "minlen.fa", "lengths.fa")
	assert files_equal(datapath('tooshort.fa'), "tooshort.tmp.fa")
	os.remove('tooshort.tmp.fa')


def test_too_short_no_primer():
	'''--too-short-output and --trim-primer'''
	run("-c -m 5 -a 330201030313112312 --trim-primer --too-short-output tooshort.tmp.fa", "minlen.noprimer.fa", "lengths.fa")
	assert files_equal(datapath('tooshort.noprimer.fa'), "tooshort.tmp.fa")
	os.remove('tooshort.tmp.fa')


def test_maximum_length():
	'''-M/--maximum-length'''
	run("-c -M 5 -a 330201030313112312", "maxlen.fa", "lengths.fa")


def test_too_long():
	'''--too-long-output'''
	run("-c -M 5 --too-long-output toolong.tmp.fa -a 330201030313112312", "maxlen.fa", "lengths.fa")
	assert files_equal(datapath('toolong.fa'), "toolong.tmp.fa")
	os.remove('toolong.tmp.fa')


def test_length_tag():
	'''454 data; -n and --length-tag'''
	run("-n 3 -e 0.1 --length-tag length= " \
		"-b TGAGACACGCAACAGGGGAAAGGCAAGGCACACAGGGGATAGG "\
		"-b TCCATCTCATCCCTGCGTGTCCCATCTGTTCCCTCCCTGTCTCA", '454.fa', '454.fa')

def test_overlap_a():
	'''-O/--overlap with -a (-c omitted on purpose)'''
	run("-O 10 -a 330201030313112312 -e 0.0 -N", "overlapa.fa", "overlapa.fa")

def test_overlap_b():
	'''-O/--overlap with -b'''
	run("-O 10 -b TTAGACATATCTCCGTCG -N", "overlapb.fa", "overlapb.fa")

def test_qualtrim():
	'''-q with low qualities'''
	run("-q 10 -a XXXXXX", "lowqual.fastq", "lowqual.fastq")

def test_qualbase():
	'''-q with low qualities, using ascii(quality+64) encoding'''
	run("-q 10 --quality-base 64 -a XXXXXX", "illumina64.fastq", "illumina64.fastq")

def test_quality_trim_only():
	'''only trim qualities, do not remove adapters'''
	run("-q 10 --quality-base 64", "illumina64.fastq", "illumina64.fastq")

def test_twoadapters():
	'''two adapters'''
	run("-a AATTTCAGGAATT -a GTTCTCTAGTTCT", "twoadapters.fasta", "twoadapters.fasta")

def test_polya():
	'''poly-A tails'''
	run("-m 24 -O 10 -a AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA", "polya.fasta", "polya.fasta")

def test_polya_brace_notation():
	'''poly-A tails'''
	run("-m 24 -O 10 -a A{35}", "polya.fasta", "polya.fasta")

def test_mask_adapter():
	'''mask adapter with N (reads maintain the same length)'''
	run("-b CAAG -n 3 --mask-adapter", "anywhere_repeat.fastq", "anywhere_repeat.fastq")

def test_gz_multiblock():
	'''compressed gz file with multiple blocks (created by concatenating two .gz files)'''
	run("-b TTAGACATATCTCCGTCG", "small.fastq", "multiblock.fastq.gz")

def test_suffix():
	'''-y/--suffix parameter, combined with _F3'''
	run("-c -e 0.12 -a 1=330201030313112312 -y _my_suffix_{name} --strip-f3", "suffix.fastq", "solid.csfasta", 'solid.qual')

def test_read_wildcard():
	'''test wildcards in reads'''
	run("--match-read-wildcards -b ACGTACGT", "wildcard.fa", "wildcard.fa")

def test_adapter_wildcard():
	'''wildcards in adapter'''
	for adapter_type, expected in (
			("-a", "wildcard_adapter.fa"),
			("-b", "wildcard_adapter_anywhere.fa")):
		with temporary_path("wildcardtmp.txt") as wildcardtmp:
			run("--wildcard-file {0} {1} ACGTNNNACGT".format(wildcardtmp, adapter_type),
				expected, "wildcard_adapter.fa")
			with open(wildcardtmp) as wct:
				lines = wct.readlines()
			lines = [ line.strip() for line in lines ]
			assert lines == ['AAA 1', 'GGG 2', 'CCC 3b', 'TTT 4b']

def test_wildcard_N():
	'''test 'N' wildcard matching with no allowed errors'''
	run("-e 0 -a GGGGGGG --match-read-wildcards", "wildcardN.fa", "wildcardN.fa")

def test_illumina_adapter_wildcard():
	run("-a VCCGAMCYUCKHRKDCUBBCNUWNSGHCGU", "illumina.fastq", "illumina.fastq.gz")

def test_adapter_front():
	'''test adapter in front'''
	run("--front ADAPTER -N", "examplefront.fa", "example.fa")

def test_literal_N():
	'''test matching literal 'N's'''
	run("-N -e 0.2 -a NNNNNNNNNNNNNN", "trimN3.fasta", "trimN3.fasta")

def test_literal_N2():
	run("-N -O 1 -g NNNNNNNNNNNNNN", "trimN5.fasta", "trimN5.fasta")

def test_literal_N_brace_notation():
	'''test matching literal 'N's'''
	run("-N -e 0.2 -a N{14}", "trimN3.fasta", "trimN3.fasta")

def test_literal_N2_brace_notation():
	run("-N -O 1 -g N{14}", "trimN5.fasta", "trimN5.fasta")

def test_anchored_front():
	run("-g ^FRONTADAPT -N", "anchored.fasta", "anchored.fasta")

def test_anchored_front_ellipsis_notation():
	run("-a FRONTADAPT... -N", "anchored.fasta", "anchored.fasta")

def test_anchored_back():
	run("-a BACKADAPTER$ -N", "anchored-back.fasta", "anchored-back.fasta")

def test_anchored_back_no_indels():
	run("-a BACKADAPTER$ -N --no-indels", "anchored-back.fasta", "anchored-back.fasta")


def test_no_indels():
	run('-a TTAGACATAT -g GAGATTGCCA --no-indels', 'no_indels.fasta', 'no_indels.fasta')


def test_issue_46():
	'''issue 46 - IndexError with --wildcard-file'''
	with temporary_path("wildcardtmp.txt") as wildcardtmp:
		run("--anywhere=AACGTN --wildcard-file={0}".format(wildcardtmp), "issue46.fasta", "issue46.fasta")

def test_strip_suffix():
	run("--strip-suffix _sequence -a XXXXXXX", "stripped.fasta", "simple.fasta")


def test_info_file():
	# The true adapter sequence in the illumina.fastq.gz data set is
	# GCCTAACTTCTTAGACTGCCTTAAGGACGT (fourth base is different)
	#
	with temporary_path("infotmp.txt") as infotmp:
		run(["--info-file", infotmp, '-a', 'adapt=GCCGAACTTCTTAGACTGCCTTAAGGACGT'], "illumina.fastq", "illumina.fastq.gz")
		assert files_equal(cutpath('illumina.info.txt'), infotmp)


def test_info_file_times():
	with temporary_path("infotmp.txt") as infotmp:
		run(["--info-file", infotmp, '--times', '2', '-a', 'adapt=GCCGAACTTCTTA', '-a', 'adapt2=GACTGCCTTAAGGACGT'], "illumina5.fastq", "illumina5.fastq")
		assert files_equal(cutpath('illumina5.info.txt'), infotmp)


def test_info_file_fasta():
	with temporary_path("infotmp.txt") as infotmp:
		# Just make sure that it runs
		run(['--info-file', infotmp, '-a', 'TTAGACATAT', '-g', 'GAGATTGCCA', '--no-indels'], 'no_indels.fasta', 'no_indels.fasta')


def test_named_adapter():
	run("-a MY_ADAPTER=GCCGAACTTCTTAGACTGCCTTAAGGACGT", "illumina.fastq", "illumina.fastq.gz")


def test_adapter_with_U():
	run("-a GCCGAACUUCUUAGACUGCCUUAAGGACGU", "illumina.fastq", "illumina.fastq.gz")


def test_no_trim():
	''' --no-trim '''
	run("--no-trim --discard-untrimmed -a CCCTAGTTAAAC", 'no-trim.fastq', 'small.fastq')


def test_bzip2():
	'''test bzip2 support'''
	run('-b TTAGACATATCTCCGTCG', 'small.fastq', 'small.fastq.bz2')


try:
	import lzma

	def test_xz():
		'''test xz support'''
		run('-b TTAGACATATCTCCGTCG', 'small.fastq', 'small.fastq.xz')
except ImportError:
	pass


@raises(SystemExit)
def test_qualfile_only():
	with redirect_stderr():
		atropos.main(['file.qual'])


@raises(SystemExit)
def test_no_args():
	with redirect_stderr():
		atropos.main([])


@raises(SystemExit)
def test_two_fastqs():
	with redirect_stderr():
		atropos.main([datapath('paired.1.fastq'), datapath('paired.2.fastq')])


def test_anchored_no_indels():
	'''anchored 5' adapter, mismatches only (no indels)'''
	run('-g ^TTAGACATAT --no-indels -e 0.1', 'anchored_no_indels.fasta', 'anchored_no_indels.fasta')


def test_anchored_no_indels_wildcard_read():
	'''anchored 5' adapter, mismatches only (no indels), but wildcards in the read count as matches'''
	run('-g ^TTAGACATAT --match-read-wildcards --no-indels -e 0.1', 'anchored_no_indels_wildcard.fasta', 'anchored_no_indels.fasta')


def test_anchored_no_indels_wildcard_adapt():
	'''anchored 5' adapter, mismatches only (no indels), but wildcards in the adapter count as matches'''
	run('-g ^TTAGACANAT --no-indels -e 0.1', 'anchored_no_indels.fasta', 'anchored_no_indels.fasta')


def test_unconditional_cut_front():
	run('-u 5', 'unconditional-front.fastq', 'small.fastq')


def test_unconditional_cut_back():
	run('-u -5', 'unconditional-back.fastq', 'small.fastq')


def test_unconditional_cut_both():
	run('-u -5 -u 5', 'unconditional-both.fastq', 'small.fastq')


def test_untrimmed_output():
	with temporary_path('untrimmed.tmp.fastq') as tmp:
		run(['-a', 'TTAGACATATCTCCGTCG', '--untrimmed-output', tmp], 'small.trimmed.fastq', 'small.fastq')
		assert files_equal(cutpath('small.untrimmed.fastq'), tmp)


def test_adapter_file():
	run('-a file:' + datapath('adapter.fasta'), 'illumina.fastq', 'illumina.fastq.gz')

def test_adapter_file_5p_anchored():
	run('-N -g file:' + datapath('prefix-adapter.fasta'), 'anchored.fasta', 'anchored.fasta')

def test_adapter_file_3p_anchored():
	run('-N -a file:' + datapath('suffix-adapter.fasta'), 'anchored-back.fasta', 'anchored-back.fasta')


def test_adapter_file_5p_anchored_no_indels():
	run('-N --no-indels -g file:' + datapath('prefix-adapter.fasta'), 'anchored.fasta', 'anchored.fasta')


def test_adapter_file_3p_anchored_no_indels():
	run('-N --no-indels -a file:' + datapath('suffix-adapter.fasta'), 'anchored-back.fasta', 'anchored-back.fasta')


def test_demultiplex():
	multiout = os.path.join(os.path.dirname(__file__), 'data', 'tmp-demulti.{name}.fasta')
	params = ['-a', 'first=AATTTCAGGAATT', '-a', 'second=GTTCTCTAGTTCT', '-o', multiout, datapath('twoadapters.fasta')]
	assert atropos.main(params) is None
	assert files_equal(cutpath('twoadapters.first.fasta'), multiout.format(name='first'))
	assert files_equal(cutpath('twoadapters.second.fasta'), multiout.format(name='second'))
	assert files_equal(cutpath('twoadapters.unknown.fasta'), multiout.format(name='unknown'))
	os.remove(multiout.format(name='first'))
	os.remove(multiout.format(name='second'))
	os.remove(multiout.format(name='unknown'))


def test_max_n():
	run('--max-n 0', 'maxn0.fasta', 'maxn.fasta')
	run('--max-n 1', 'maxn1.fasta', 'maxn.fasta')
	run('--max-n 2', 'maxn2.fasta', 'maxn.fasta')
	run('--max-n 0.2', 'maxn0.2.fasta', 'maxn.fasta')
	run('--max-n 0.4', 'maxn0.4.fasta', 'maxn.fasta')


def test_quiet_is_quiet():
	captured_standard_output = StringIO()
	captured_standard_error = StringIO()
	try:
		old_stdout = sys.stdout
		old_stderr = sys.stderr
		sys.stdout = captured_standard_output
		sys.stderr = captured_standard_error
		atropos.main(['-o', '/dev/null', '--quiet', '-a', 'XXXX', datapath('illumina.fastq.gz')])
	finally:
		sys.stdout = old_stdout
		sys.stderr = old_stderr
	assert captured_standard_output.getvalue() == ''
	assert captured_standard_error.getvalue() == ''


def test_nextseq():
	run('--nextseq-trim 22', 'nextseq.fastq', 'nextseq.fastq')


def test_linked():
	run('-a AAAAAAAAAA...TTTTTTTTTT', 'linked.fasta', 'linked.fasta')


def test_fasta():
	run('-a TTAGACATATCTCCGTCG', 'small.fasta', 'small.fastq')

def test_custom_bisulfite_1():
	run('-b TTAGACATATCTCCGTCG -q 0,0 --bisulfite 2,2,1,1', 'small.fastq', 'small.fastq')
def test_custom_bisulfite_2():
	run('-b TTAGACATATCTCCGTCG -q 0,0 --bisulfite 15,15,1,1', 'small_mincut1.fastq', 'small.fastq')
def test_custom_bisulfite_3():
	run('-b TTAGACATATCTCCGTCG -q 0,0 --bisulfite 2,2,1,0', 'small_mincut2.fastq', 'small.fastq')
def test_custom_bisulfite_4():
	run('-b TTAGACATATCTCCGTCG -q 0,0 --bisulfite 2,2,0,0', 'small_mincut3.fastq', 'small.fastq')



# coding: utf-8
from __future__ import print_function, division, absolute_import

from atropos.seqio import ColorspaceSequence, Sequence
from atropos.adapters import Adapter, ColorspaceAdapter, PREFIX, BACK
from atropos.modifiers import AdapterCutter

def test_cs_5p():
	read = ColorspaceSequence("name", "0123", "DEFG", "T")
	adapter = ColorspaceAdapter("CG", PREFIX, 0.1)
	cutter = AdapterCutter([adapter])
	trimmed_read = cutter(read)
	# no assertion here, just make sure the above code runs without
	# an exception


def test_statistics():
	read = Sequence('name', 'AAAACCCCAAAA')
	adapters = [Adapter('CCCC', BACK, 0.1)]
	cutter = AdapterCutter(adapters, times=3)
	trimmed_read = cutter(read)
	# TODO make this a lot simpler
	trimmed_bp = 0
	for adapter in adapters:
		for d in (adapter.lengths_front, adapter.lengths_back):
			trimmed_bp += sum(seqlen * count for (seqlen, count) in d.items())
	assert trimmed_bp <= len(read), trimmed_bp



# kate: syntax Python;
# cython: profile=False, emit_code_comments=False
from __future__ import print_function, division, absolute_import
from .xopen import xopen
from .seqio import _shorten, FormatError, SequenceReader

cdef class Sequence(object):
	"""
	A record in a FASTQ file. Also used for FASTA (then the qualities attribute
	is None). qualities is a string and it contains the qualities encoded as
	ascii(qual+33).

	If an adapter has been matched to the sequence, the 'match' attribute is
	set to the corresponding Match instance.
	"""
	cdef:
		public str name
		public str sequence
		public str qualities
		public str name2
		public int original_length
		public object match
		public object match_info
		public object clipped
		public bint merged

	def __init__(self, str name, str sequence, str qualities=None, str name2='',
				 original_length=None, match=None, match_info=None, clipped=None,
				 merged=False):
		"""Set qualities to None if there are no quality values"""
		self.name = name
		self.sequence = sequence
		self.qualities = qualities
		self.name2 = name2
		self.original_length = original_length or len(sequence)
		self.match = match
		self.match_info = match_info
		self.clipped = clipped or [0,0,0,0]
		self.merged = merged
		if qualities is not None and len(qualities) != len(sequence):
			rname = _shorten(name)
			raise FormatError("In read named {0!r}: length of quality sequence ({1}) and length "
				"of read ({2}) do not match".format(
					rname, len(qualities), len(sequence)))
	
	def subseq(self, begin=0, end=None):
		if end is None:
			new_read = self[begin:]
		else:
			new_read = self[begin:end]
		end_bases = len(self) - end
		offset = 2 if self.match else 0
		if begin:
			new_read.clipped[offset] += begin
		if end_bases:
			new_read.clipped[offset+1] += end_bases
		return (begin, end_bases, new_read)
	
	def clip(self, front=0, back=0):
		if back < 0:
			new_read = self[front:back]
			back *= -1
		else:
			new_read = self[front:]
		offset = 2 if self.match else 0
		if front:
			new_read.clipped[offset] += front
		if back:
			new_read.clipped[offset+1] += back
		return (front, back, new_read)
	
	def __getitem__(self, key):
		"""slicing"""
		return self.__class__(
			self.name,
			self.sequence[key],
			self.qualities[key] if self.qualities is not None else None,
			self.name2,
			self.original_length,
			self.match,
			self.match_info,
			list(self.clipped),
			self.merged
		)

	def __repr__(self):
		qstr = ''
		if self.qualities is not None:
			qstr = ', qualities={0!r}'.format(_shorten(self.qualities))
		return '<Sequence(name={0!r}, sequence={1!r}{2})>'.format(_shorten(self.name), _shorten(self.sequence), qstr)

	def __len__(self):
		return len(self.sequence)

	def __richcmp__(self, other, int op):
		if 2 <= op <= 3:
			eq = self.name == other.name and \
				self.sequence == other.sequence and \
				self.qualities == other.qualities
			if op == 2:
				return eq
			else:
				return not eq
		else:
			raise NotImplementedError()

	def __reduce__(self):
		return (Sequence, (self.name, self.sequence, self.qualities, self.name2))


class FastqReader(SequenceReader):
	"""
	Reader for FASTQ files. Does not support multi-line FASTQ files.
	"""
	def __init__(self, file, sequence_class=Sequence):
		"""
		file is a filename or a file-like object.
		If file is a filename, then .gz files are supported.
		"""
		super(FastqReader, self).__init__(file)
		self.sequence_class = sequence_class
		self.delivers_qualities = True

	def __iter__(self):
		"""
		Yield Sequence objects
		"""
		cdef int i = 0
		cdef int strip
		cdef str line, name, qualities, sequence, name2
		sequence_class = self.sequence_class

		it = iter(self._file)
		line = next(it)
		if not (line and line[0] == '@'):
			raise FormatError("Line {0} in FASTQ file is expected to start with '@', but found {1!r}".format(i+1, line[:10]))
		strip = -2 if line.endswith('\r\n') else -1
		name = line[1:strip]

		i = 1
		for line in it:
			if i == 0:
				if not (line and line[0] == '@'):
					raise FormatError("Line {0} in FASTQ file is expected to start with '@', but found {1!r}".format(i+1, line[:10]))
				name = line[1:strip]
			elif i == 1:
				sequence = line[:strip]
			elif i == 2:
				if line == '+\n':  # check most common case first
					name2 = ''
				else:
					line = line[:strip]
					if not (line and line[0] == '+'):
						raise FormatError("Line {0} in FASTQ file is expected to start with '+', but found {1!r}".format(i+1, line[:10]))
					if len(line) > 1:
						if not line[1:] == name:
							raise FormatError(
								"At line {0}: Sequence descriptions in the FASTQ file don't match "
								"({1!r} != {2!r}).\n"
								"The second sequence description must be either empty "
								"or equal to the first description.".format(i+1,
									name, line[1:]))
						name2 = name
					else:
						name2 = ''
			elif i == 3:
				if len(line) == len(sequence) - strip:
					qualities = line[:strip]
				else:
					qualities = line.rstrip('\r\n')
				yield sequence_class(name, sequence, qualities, name2=name2)
			i = (i + 1) % 4
		if i != 0:
			raise FormatError("FASTQ file ended prematurely")



# coding: utf-8
"""
This module implements all the read modifications that atropos supports.
A modifier must be callable. It is implemented as a function if no parameters
need to be stored, and as a class with a __call__ method if there are parameters
(or statistics).
"""
import re
from collections import OrderedDict
from atropos.qualtrim import quality_trim_index, nextseq_trim_index
from atropos.align import Aligner, SEMIGLOBAL
import logging

class AdapterCutter(object):
    """
    Repeatedly find one of multiple adapters in reads.
    The number of times the search is repeated is specified by the
    times parameter.
    """

    def __init__(self, adapters=[], times=1, action='trim', keep_match_info=False):
        """
        adapters -- list of Adapter objects

        action -- What to do with a found adapter: None, 'trim', or 'mask'
        """
        super(AdapterCutter, self).__init__()
        self.adapters = adapters
        self.times = times
        self.action = action
        self.with_adapters = 0
        self.keep_match_info = keep_match_info

    def _best_match(self, read):
        """
        Find the best matching adapter in the given read.

        Return either a Match instance or None if there are no matches.
        """
        best = None
        for adapter in self.adapters:
            match = adapter.match_to(read)
            if match is None:
                continue

            # the no. of matches determines which adapter fits best
            if best is None or match.matches > best.matches:
                best = match
        return best

    def __call__(self, read):
        """
        Determine the adapter that best matches the given read.
        Since the best adapter is searched repeatedly, a list
        of Match instances is returned, which
        need to be applied consecutively to the read.
        The list is empty if there are no adapter matches.

        The read is converted to uppercase before it is compared to the adapter
        sequences.

        Cut found adapters from a single read. Return modified read.
        """
        matches = []

        # try at most self.times times to remove an adapter
        trimmed_read = read
        for t in range(self.times):
            match = self._best_match(trimmed_read)
            if match is None:
                # nothing found
                break
            matches.append(match)
            trimmed_read = match.adapter.trimmed(match)
        
        if not matches:
            trimmed_read.match = None
            trimmed_read.match_info = None
            return trimmed_read
        
        if __debug__:
            assert len(trimmed_read) < len(read), "Trimmed read isn't shorter than original"
        
        if self.action == 'trim':
            # read is already trimmed, nothing to do
            pass
        elif self.action == 'mask':
            # add N from last modification
            masked_sequence = trimmed_read.sequence
            for match in sorted(matches, reverse=True, key=lambda m: m.astart):
                ns = 'N' * (len(match.read.sequence) -
                            len(match.adapter.trimmed(match).sequence))
                # add N depending on match position
                if match.front:
                    masked_sequence = ns + masked_sequence
                else:
                    masked_sequence += ns
            # set masked sequence as sequence with original quality
            trimmed_read.sequence = masked_sequence
            trimmed_read.qualities = matches[0].read.qualities

            assert len(trimmed_read.sequence) == len(read)
        elif self.action is None:
            trimmed_read = read
        
        trimmed_read.match = matches[-1]
        trimmed_read.match_info = [match.get_info_record() for match in matches]
        
        self.with_adapters += 1
        return trimmed_read

class Trimmer(object):
    def __init__(self):
        self.trimmed_bases = 0
    
    def __call__(self, read):
        raise NotImplemented()
    
    def subseq(self, read, begin=0, end=None):
        if begin or (end is not None):
            front_bases, back_bases, new_read = read.subseq(begin, end)
            self.trimmed_bases += front_bases + back_bases
            return new_read
        else:
            return read
    
    def clip(self, read, front=0, back=0):
        if front or back:
            front_bases, back_bases, new_read = read.clip(front, back)
            self.trimmed_bases += front_bases + back_bases
            return new_read
        else:
            return read

class UnconditionalCutter(Trimmer):
    """
    A modifier that unconditionally removes the first n or the last n bases from a read.
    Equivalent to MinCutter with count_trimmed=False and only_trimmed=False, but
    UnconditionalCutter is applied before adapter trimming.
    
    If the length is positive, the bases are removed from the beginning of the read.
    If the length is negative, the bases are removed from the end of the read.
    """
    
    display_str = "Cut unconditionally"
    
    def __init__(self, lengths=[]):
        super(UnconditionalCutter, self).__init__()
        self.front_length = sum(l for l in lengths if l > 0)
        self.back_length = sum(l for l in lengths if l < 0)

    def __call__(self, read):
        return self.clip(read, self.front_length, self.back_length)

class MinCutter(Trimmer):
    """
    Ensure that a minimum number of bases have been trimmed off each end.
    count_trimmed :: whether to consider bases cut before or during adapter trimming
        when counting the number of bases that have already been cut
    only_trimmed :: only cut read ends if they have already been adapter-trimmed.
    """
    
    display_str = "Cut conditionally"
    
    def __init__(self, lengths=[], count_trimmed=True, only_trimmed=False):
        super(MinCutter, self).__init__()
        self.front_length = sum(l for l in lengths if l > 0)
        self.back_length = sum(l for l in lengths if l < 0)
        self.count_trimmed = count_trimmed
        self.only_trimmed = only_trimmed
    
    def __call__(self, read):
        trim_front = trim_back = True
        if self.only_trimmed:
            if read.match:
                is_front = [match.is_front for match in read.match_info]
                if not any(is_front):
                    trim_front = False
                elif all(is_front):
                    trim_back = False
            else:
                return read
        
        def to_trim(offset, is_front):
            if self.count_trimmed:
                trimmed = read.clipped[offset] + read.clipped[offset+2]
                if read.match:
                    trimmed += sum(i.rsize_total for i in read.match_info if is_front == i.is_front)
            elif read.match:
                trimmed = read.clipped[offset+2]
            else:
                trimmed = read.clipped[offset]
            
            if is_front:
                return max(self.front_length - trimmed, 0)
            else:
                return min(trimmed + self.back_length, 0)
        
        return self.clip(
            read,
            to_trim(0, True) if trim_front else 0,
            to_trim(1, False) if trim_back else 0
        )

class LengthTagModifier(object):
    """
    Replace "length=..." strings in read names.
    """
    def __init__(self, length_tag="length="):
        self.regex = re.compile(r"\b" + length_tag + r"[0-9]*\b")
        self.length_tag = length_tag

    def __call__(self, read):
        read = read[:]
        if read.name.find(self.length_tag) >= 0:
            read.name = self.regex.sub(self.length_tag + str(len(read.sequence)), read.name)
        return read

class SuffixRemover(object):
    """
    Remove a given suffix from read names.
    """
    def __init__(self, suffixes=[]):
        self.suffixes = suffixes

    def __call__(self, read):
        name = read.name
        for s in self.suffixes:
            if name.endswith(s):
                name = name[:-len(s)]
        read = read[:]
        read.name = name
        return read

class PrefixSuffixAdder(object):
    """
    Add a suffix and a prefix to read names
    """
    def __init__(self, prefix="", suffix=""):
        self.prefix = prefix
        self.suffix = suffix

    def __call__(self, read):
        read = read[:]
        adapter_name = 'no_adapter' if read.match is None else read.match.adapter.name
        read.name = self.prefix.replace('{name}', adapter_name) + read.name + \
            self.suffix.replace('{name}', adapter_name)
        return read

class DoubleEncoder(object):
    """
    Double-encode colorspace reads, using characters ACGTN to represent colors.
    """
    def __init__(self):
        self.double_encode_trans = str.maketrans('0123.', 'ACGTN')

    def __call__(self, read):
        read = read[:]
        read.sequence = read.sequence.translate(self.double_encode_trans)
        return read

class ZeroCapper(object):
    """
    Change negative quality values of a read to zero
    """
    def __init__(self, quality_base=33):
        qb = quality_base
        self.zero_cap_trans = str.maketrans(''.join(map(chr, range(qb))), chr(qb) * qb)

    def __call__(self, read):
        read = read[:]
        read.qualities = read.qualities.translate(self.zero_cap_trans)
        return read

class PrimerTrimmer(Trimmer):
    display_str = "Primer-trimmed"
    
    def __call__(self, read):
        """Trim primer base from colorspace reads"""
        read = self.clip(read, 1)
        read.primer = ''
        return read

class NextseqQualityTrimmer(Trimmer):
    display_str = "Quality trimmed (NextSeq)"
    
    def __init__(self, cutoff=0, base=33):
        super(NextseqQualityTrimmer, self).__init__()
        self.cutoff = cutoff
        self.base = base
    
    def __call__(self, read):
        stop = nextseq_trim_index(read, self.cutoff, self.base)
        return self.subseq(read, end=stop)

class QualityTrimmer(Trimmer):
    display_str = "Quality-trimmed"
    
    def __init__(self, cutoff_front=0, cutoff_back=0, base=33):
        super(QualityTrimmer, self).__init__()
        self.cutoff_front = cutoff_front
        self.cutoff_back = cutoff_back
        self.base = base

    def __call__(self, read):
        start, stop = quality_trim_index(read.qualities, self.cutoff_front, self.cutoff_back, self.base)
        return self.subseq(read, start, stop)

class NEndTrimmer(Trimmer):
    """Trims Ns from the 3' and 5' end of reads"""
    
    display_str = "End Ns trimmed"
    
    def __init__(self):
        super(NEndTrimmer, self).__init__()
        self.start_trim = re.compile(r'^N+')
        self.end_trim = re.compile(r'N+$')

    def __call__(self, read):
        sequence = read.sequence
        start_cut = self.start_trim.match(sequence)
        end_cut = self.end_trim.search(sequence)
        start_cut = start_cut.end() if start_cut else 0
        end_cut = end_cut.start() if end_cut else len(read)
        return self.subseq(read, start_cut, end_cut)

class BisulfiteTrimmer(MinCutter):
    """
    Generic bisulfite trimmer. Sequences that are adapter-trimmed are further
    trimmed 2 bp on the 3' end to remove potential methylation-biased bases
    from the end-repair reaction.
    """
    
    display_str = "Bisulfite-trimmed"
    
    def __init__(self, trim_5p=0, trim_3p=2):
        super(BisulfiteTrimmer, self).__init__((trim_5p, -1 * trim_3p), count_trimmed=False, only_trimmed=True)

class NonDirectionalBisulfiteTrimmer(object):
    """
    For non-directional RRBS/WGBS libraries (which implies that they were digested
    using MspI), sequences starting with either 'CAA' or 'CGA' will have 2 bp
    trimmed off either end to remove potential methylation-biased bases from the
    end-repair reaction. Otherwise only the 3' end of adapter-trimmed reads are
    further trimmed 2 bp.
    """
    
    display_str = "Bisulfite-trimmed (Non-directional)"
    _regex = re.compile("^C[AG]A")
    
    def __init__(self, trim_5p=2, trim_3p=2):
        self._3pCutter = MinCutter((-1 * trim_3p,), count_trimmed=False, only_trimmed=True)
        self._bothCutter = MinCutter((trim_5p, -1 * trim_3p), count_trimmed=False)
    
    def __call__(self, read):
        cutter = None
        if self._regex.match(read.sequence):
            cutter = self._bothCutter
        elif read.match:
            cutter = self._3pCutter
        return cutter(read) if cutter else read
    
    @property
    def modified_bases(self):
        return self._3pCutter.trimmed_bases + self._bothCutter.trimmed_bases

class EpiGnomeBisulfiteTrimmer(MinCutter):
    """EpiGnome reads are trimmed 6 bp on the 5' end."""
    
    display_str = "Bisulfite-trimmed (EpiGnome)"
    
    def __init__(self):
        super(EpiGnomeBisulfiteTrimmer, self).__init__((6,), count_trimmed=True)

class SwiftBisulfiteTrimmer(object):
    """
    For WGBS libraries prepared with the Swift Accel-NGS kit, 10 bp are trimmed off the 5' end.
    Additionally, if the read length is longer than 100 bp, 10 bp ar trimmed off the 3' end.
    """
    
    display_str = "Bisulfite-trimmed (Swift)"
    
    def __init__(self):
        self._shortCutter = MinCutter((10,), count_trimmed=False, only_trimmed=False)
        self._longCutter = MinCutter((10,-10), count_trimmed=False, only_trimmed=False)
    
    def __call__(self, read):
        cutter = self._longCutter if read.original_length > 100 else self._shortCutter
        return cutter(read)

    @property
    def modified_bases(self):
        return self._shortCutter.trimmed_bases + self._longCutter.trimmed_bases

complement = {
    'A' : 'T',
    'C' : 'G',
    'R' : 'Y',
    'S' : 'S',
    'W' : 'W',
    'K' : 'M',
    'B' : 'V',
    'D' : 'H',
    'N' : 'N'
}
for k,v in list(complement.items()):
    complement[v] = k
    complement[k.lower()] = v.lower()
    complement[v.lower()] = k.lower()

def reverse_complement(seq):
    return "".join(complement[base] for base in reversed(seq))

class MergeOverlapping(object):
    def __init__(self, min_overlap=0.9, error_rate=0.1):
        self.min_overlap = int(min_overlap) if min_overlap > 1 else min_overlap
        self.error_rate = error_rate
    
    def __call__(self, read1, read2):
        if read1.match and not read1.match.front and read2.match and not read2.match.front:
            min_overlap = self.min_overlap
            l1 = len(read1.sequence)
            l2 = len(read2.sequence)
            if min_overlap <= 1:
                min_overlap = max(2, round(self.min_overlap * min(l1, l2)))
                
            aligner = Aligner(read1.sequence, self.error_rate, SEMIGLOBAL)
            read2_rc = reverse_complement(read2.sequence)
            # TODO: Resolve overlap. We offer the user the choice to set mismatching bases to
            # the one with the highest quality or to 'N'. Additionally, set qualities of overlapping
            # bases to the highest value. The acutal alignment is not returned, so if there are
            # mismatches we don't know what they are, and figuring it out is complicated by indels.
            alignment = aligner.locate(read2_rc)
            if alignment:
                r1_start, r1_stop, r2_start, r2_stop, matches, errors = alignment
                if matches >= min_overlap:
                    if r2_start == 0 and r2_stop == l2:
                        # r2 is fully contained in r1
                        pass
                    elif r1_start == 0 and r1_stop == l1:
                        # r1 is fully contained in r2
                        read1.sequence = read2_rc
                        read1.qualities = "".join(reversed(read2.qualities))
                    elif r1_start > 0:
                        read1.sequence += read2_rc[r2_stop:]
                        if read1.qualities and read2.qualities:
                            read1.qualities += "".join(reversed(read2.qualities))[r2_stop:]
                    elif r2_start > 0:
                        read1.sequence = read2_rc + read1[r1_stop:]
                        if read1.qualities and read2.qualities:
                            read1.qualities = "".join(reversed(read2.qualities)) + read1.qualities[r1_stop:]
                    else:
                        raise Exception(
                            "Invalid alignment while trying to merge read {}: {}".format(
                                read1.name, ",".join(str(i) for i in alignment)))
                    
                    read1.merged = True
                    read2 = None
                
        return (read1, read2)

class Modifiers(object):
    def __init__(self, paired, merger=None):
        self.modifiers1 = OrderedDict()
        self.modifiers2 = OrderedDict()
        self.paired = paired
        self.merger = merger
    
    def add_modifier(self, mod_class, read=1|2, **kwargs):
        if read & 1 > 0:
            self.modifiers1[mod_class] = mod_class(**kwargs)
        if read & 2 > 0 and self.paired == "both":
            self.modifiers2[mod_class] = mod_class(**kwargs)
    
    def get_modifier_pair(self, mod_class):
        return [self.modifiers1.get(mod_class, None), self.modifiers2.get(mod_class, None)]
    
    def modify(self, record):
        bp = [0,0]
        if self.paired:
            read1, read2 = record
            bp[0] = len(read1.sequence)
            bp[1] = len(read2.sequence)
            for mod in self.modifiers1.values():
                read1 = mod(read1)
            for mod in self.modifiers2.values():
                read2 = mod(read2)
            if self.merger:
                read1, read2 = self.merger.merge(read1, read2)
            reads = [read1, read2]
        else:
            read = record
            bp[0] = len(read.sequence)
            for mod in self.modifiers1.values():
                read = mod(read)
            reads = [read]
        return (reads, bp)
    
    def __getitem__(self, i):
        if i == 0:
            return self.modifiers1
        else:
            return self.modifiers2
    
    def get_trimmer_classes(self):
        return (set(k for k,v in self.modifiers1.items() if isinstance(v, Trimmer)) |
                set(k for k,v in self.modifiers2.items() if isinstance(v, Trimmer)))



# coding: utf-8
import gzip
import os
import random
import sys
from nose.tools import raises
from atropos.xopen import xopen, lzma, open_output, get_compressor
from .utils import temporary_path

base = "tests/data/small.fastq"
files = [ base + ext for ext in ['', '.gz', '.bz2' ] ]
if lzma is not None:
    files.append(base + '.xz')

def test_context_manager():
    major, minor = sys.version_info[0:2]
    for name in files:
        if major == 2 and minor == 6:
            continue  # Py26 compression libraries do not support context manager protocol.
        with xopen(name, 'rt') as f:
            lines = list(f)
            assert len(lines) == 12
            assert lines[5] == 'AGCCGCTANGACGGGTTGGCCCTTAGACGTATCT\n', name
            f.close()

def test_append():
    for ext in ["", ".gz"]:  # BZ2 does NOT support append
        text = "AB"
        reference = text + text
        filename = 'truncated.fastq' + ext
        mode = 'a'
        if ext != "":
            mode = 'ab'
            text = text.encode()
            reference = text + text
            text = get_compressor(filename).compress(text)  # On Py3, need to send BYTES, not unicode
        print("Trying ext=%s" % ext)
        with temporary_path(filename) as path:
            try:
                os.unlink(path)
            except OSError:
                pass
            with open_output(path, mode) as f:
                f.write(text)
            print(path)
            with open_output(path, mode) as f:
                f.write(text)
            with xopen(path, 'r') as f:
                try:
                    reference = reference.decode("utf-8")
                except AttributeError:
                    pass
                for appended in f:
                	assert appended == reference

def test_xopen_text():
    for name in files:
        f = xopen(name, 'rt')
        lines = list(f)
        assert len(lines) == 12
        assert lines[5] == 'AGCCGCTANGACGGGTTGGCCCTTAGACGTATCT\n', name
        f.close()


def test_xopen_binary():
    for name in files:
        f = xopen(name, 'rb')
        lines = list(f)
        assert len(lines) == 12
        assert lines[5] == b'AGCCGCTANGACGGGTTGGCCCTTAGACGTATCT\n', name
        f.close()



# coding: utf-8
from __future__ import print_function, division, absolute_import

from collections import defaultdict
import random
import sys
import os
from io import StringIO
import shutil
from textwrap import dedent
from nose.tools import raises
from tempfile import mkdtemp
from atropos.seqio import (Sequence, ColorspaceSequence, FormatError,
	FastaReader, FastqReader, FastaQualReader, InterleavedSequenceReader,
	FastaFormat, FastqFormat, InterleavedFormatter, get_format,
	open_reader as openseq, sequence_names_match, open_output)
from atropos.xopen import xopen, open_output
from .utils import temporary_path

# files tests/data/simple.fast{q,a}
simple_fastq = [
	Sequence("first_sequence", "SEQUENCE1", ":6;;8<=:<"),
	Sequence("second_sequence", "SEQUENCE2", "83<??:(61")
	]

simple_fasta = [ Sequence(x.name, x.sequence, None) for x in simple_fastq ]


class TestSequence:
	@raises(FormatError)
	def test_too_many_qualities(self):
		Sequence(name="name", sequence="ACGT", qualities="#####")

	@raises(FormatError)
	def test_too_many_qualities_colorspace(self):
		ColorspaceSequence(name="name", sequence="T0123", qualities="#####")

	@raises(FormatError)
	def test_invalid_primer(self):
		ColorspaceSequence(name="name", sequence="K0123", qualities="####")


class TestFastaReader:
	def test(self):
		with FastaReader("tests/data/simple.fasta") as f:
			reads = list(f)
		assert reads == simple_fasta

		fasta = StringIO(">first_sequence\nSEQUENCE1\n>second_sequence\nSEQUENCE2\n")
		reads = list(FastaReader(fasta))
		assert reads == simple_fasta

	def test_with_comments(self):
		fasta = StringIO(dedent(
			"""
			# a comment
			# another one
			>first_sequence
			SEQUENCE1
			>second_sequence
			SEQUENCE2
			"""))
		reads = list(FastaReader(fasta))
		assert reads == simple_fasta

	@raises(FormatError)
	def test_wrong_format(self):
		fasta = StringIO(dedent(
			"""
			# a comment
			# another one
			unexpected
			>first_sequence
			SEQUENCE1
			>second_sequence
			SEQUENCE2
			"""))
		reads = list(FastaReader(fasta))

	def test_fastareader_keeplinebreaks(self):
		with FastaReader("tests/data/simple.fasta", keep_linebreaks=True) as f:
			reads = list(f)
		assert reads[0] == simple_fasta[0]
		assert reads[1].sequence == 'SEQUEN\nCE2'

	def test_context_manager(self):
		filename = "tests/data/simple.fasta"
		with open(filename) as f:
			assert not f.closed
			reads = list(openseq(f))
			assert not f.closed
		assert f.closed

		with FastaReader(filename) as sr:
			tmp_sr = sr
			assert not sr._file.closed
			reads = list(sr)
			assert not sr._file.closed
		assert tmp_sr._file is None
		# Open it a second time
		with FastaReader(filename) as sr:
			pass


class TestFastqReader:
	def test_fastqreader(self):
		with FastqReader("tests/data/simple.fastq") as f:
			reads = list(f)
		assert reads == simple_fastq

	def test_fastqreader_dos(self):
		with FastqReader("tests/data/dos.fastq") as f:
			dos_reads = list(f)
		with FastqReader("tests/data/small.fastq") as f:
			unix_reads = list(f)
		assert dos_reads == unix_reads

	@raises(FormatError)
	def test_fastq_wrongformat(self):
		with FastqReader("tests/data/withplus.fastq") as f:
			reads = list(f)

	@raises(FormatError)
	def test_fastq_incomplete(self):
		fastq = StringIO("@name\nACGT+\n")
		with FastqReader(fastq) as fq:
			list(fq)

	def test_context_manager(self):
		filename = "tests/data/simple.fastq"
		with open(filename) as f:
			assert not f.closed
			reads = list(openseq(f))
			assert not f.closed
		assert f.closed

		with FastqReader(filename) as sr:
			tmp_sr = sr
			assert not sr._file.closed
			reads = list(sr)
			assert not sr._file.closed
		assert tmp_sr._file is None


class TestFastaQualReader:
	@raises(FormatError)
	def test_mismatching_read_names(self):
		fasta = StringIO(">name\nACG")
		qual = StringIO(">nome\n3 5 7")
		list(FastaQualReader(fasta, qual))

	@raises(FormatError)
	def test_invalid_quality_value(self):
		fasta = StringIO(">name\nACG")
		qual = StringIO(">name\n3 xx 7")
		list(FastaQualReader(fasta, qual))


class TestSeqioOpen:
	def setup(self):
		self._tmpdir = mkdtemp()

	def teardown(self):
		shutil.rmtree(self._tmpdir)

	def test_sequence_reader(self):
		# test the autodetection
		with openseq("tests/data/simple.fastq") as f:
			reads = list(f)
		assert reads == simple_fastq

		with openseq("tests/data/simple.fasta") as f:
			reads = list(f)
		assert reads == simple_fasta

		with open("tests/data/simple.fastq") as f:
			reads = list(openseq(f))
		assert reads == simple_fastq

		# make the name attribute unavailable
		f = StringIO(open("tests/data/simple.fastq").read())
		reads = list(openseq(f))
		assert reads == simple_fastq

		f = StringIO(open("tests/data/simple.fasta").read())
		reads = list(openseq(f))
		assert reads == simple_fasta

	def test_autodetect_fasta_format(self):
		path = os.path.join(self._tmpdir, 'tmp.fasta')
		fmt = get_format(path)
		assert isinstance(fmt, FastaFormat)
		with open_output(path, "w") as f:
			for seq in simple_fasta:
				f.write(fmt.format(seq))
		assert list(openseq(path)) == simple_fasta

	def test_write_qualities_to_fasta(self):
		path = os.path.join(self._tmpdir, 'tmp.fasta')
		fmt = get_format(path, qualities=True)
		assert isinstance(fmt, FastaFormat)
		with open_output(path, "w") as f:
			for seq in simple_fastq:
				f.write(fmt.format(seq))
		assert list(openseq(path)) == simple_fasta

	def test_autodetect_fastq_format(self):
		path = os.path.join(self._tmpdir, 'tmp.fastq')
		fmt = get_format(path)
		with open_output(path, "w") as f:
			for seq in simple_fastq:
				f.write(fmt.format(seq))
		assert list(openseq(path)) == simple_fastq

	@raises(ValueError)
	def test_fastq_qualities_missing(self):
		path = os.path.join(self._tmpdir, 'tmp.fastq')
		get_format(path, qualities=False)


class TestInterleavedReader:
	def test(self):
		expected = [
			(Sequence('read1/1 some text', 'TTATTTGTCTCCAGC', '##HHHHHHHHHHHHH'),
			Sequence('read1/2 other text', 'GCTGGAGACAAATAA', 'HHHHHHHHHHHHHHH')),
			(Sequence('read3/1', 'CCAACTTGATATTAATAACA', 'HHHHHHHHHHHHHHHHHHHH'),
			Sequence('read3/2', 'TGTTATTAATATCAAGTTGG', '#HHHHHHHHHHHHHHHHHHH'))
		]
		reads = list(InterleavedSequenceReader("tests/cut/interleaved.fastq"))
		for (r1, r2), (e1, e2) in zip(reads, expected):
			print(r1, r2, e1, e2)

		assert reads == expected
		with openseq("tests/cut/interleaved.fastq", interleaved=True) as f:
			reads = list(f)
		assert reads == expected

	@raises(FormatError)
	def test_missing_partner(self):
		s = StringIO('@r1\nACG\n+\nHHH')
		list(InterleavedSequenceReader(s))

	@raises(FormatError)
	def test_incorrectly_paired(self):
		s = StringIO('@r1/1\nACG\n+\nHHH\n@wrong_name\nTTT\n+\nHHH')
		list(InterleavedSequenceReader(s))


class TestFastaWriter:
	def setup(self):
		self._tmpdir = mkdtemp()
		self.path = os.path.join(self._tmpdir, 'tmp.fasta')

	def teardown(self):
		shutil.rmtree(self._tmpdir)

	def test(self):
		fmt = FastaFormat()
		with open_output(self.path, "w") as fw:
			fw.write(fmt.format_entry("name", "CCATA"))
			fw.write(fmt.format_entry("name2", "HELLO"))
		with open(self.path) as t:
			assert t.read() == '>name\nCCATA\n>name2\nHELLO\n'

	def test_linelength(self):
		fmt = FastaFormat(line_length=3)
		with open_output(self.path, "w") as fw:
			fw.write(fmt.format_entry("r1", "ACG"))
			fw.write(fmt.format_entry("r2", "CCAT"))
			fw.write(fmt.format_entry("r3", "TACCAG"))
		with open(self.path) as t:
			x=t.read()
			print(x)
			assert x == '>r1\nACG\n>r2\nCCA\nT\n>r3\nTAC\nCAG\n'

	def test_write_sequence_object(self):
		fmt = FastaFormat()
		with open_output(self.path, "w") as fw:
			fw.write(fmt.format(Sequence("name", "CCATA")))
			fw.write(fmt.format(Sequence("name2", "HELLO")))
		with open(self.path) as t:
			assert t.read() == '>name\nCCATA\n>name2\nHELLO\n'

	def test_write_zero_length_sequence(self):
		fmt = FastaFormat()
		s = fmt.format_entry("name", "")
		assert s == '>name\n\n', '{0!r}'.format(s)


class TestFastqWriter:
	def setup(self):
		self._tmpdir = mkdtemp()
		self.path = os.path.join(self._tmpdir, 'tmp.fastq')

	def teardown(self):
		shutil.rmtree(self._tmpdir)

	def test(self):
		fmt = FastqFormat()
		with open_output(self.path, "w") as fw:
			fw.write(fmt.format_entry("name", "CCATA", "!#!#!"))
			fw.write(fmt.format_entry("name2", "HELLO", "&&&!&&"))
		with open(self.path) as t:
			assert t.read() == '@name\nCCATA\n+\n!#!#!\n@name2\nHELLO\n+\n&&&!&&\n'

	def test_twoheaders(self):
		fmt = FastqFormat()
		with open_output(self.path, "w") as fw:
			fw.write(fmt.format(Sequence("name", "CCATA", "!#!#!", name2="name")))
			fw.write(fmt.format(Sequence("name2", "HELLO", "&&&!&", name2="name2")))
		with open(self.path) as t:
			assert t.read() == '@name\nCCATA\n+name\n!#!#!\n@name2\nHELLO\n+name2\n&&&!&\n'


class TestInterleavedWriter:
	def test(self):
		reads = [
			(Sequence('A/1 comment', 'TTA', '##H'),
			Sequence('A/2 comment', 'GCT', 'HH#')),
			(Sequence('B/1', 'CC', 'HH'),
			Sequence('B/2', 'TG', '#H'))
		]
		fmt = InterleavedFormatter(FastqFormat(), "foo")
		result = defaultdict(lambda: [])
		for read1, read2 in reads:
			fmt.format(result, read1, read2)
		assert fmt.written == 2
		assert fmt.read1_bp == 5
		assert fmt.read2_bp == 5
		assert "foo" in result
		assert "".join(result["foo"]) == '@A/1 comment\nTTA\n+\n##H\n@A/2 comment\nGCT\n+\nHH#\n@B/1\nCC\n+\nHH\n@B/2\nTG\n+\n#H\n'

class TestPairedSequenceReader:
	def test_sequence_names_match(self):
		def match(name1, name2):
			seq1 = Sequence(name1, 'ACGT')
			seq2 = Sequence(name2, 'AACC')
			return sequence_names_match(seq1, seq2)

		assert match('abc', 'abc')
		assert match('abc/1', 'abc/2')
		assert match('abc.1', 'abc.2')
		assert match('abc1', 'abc2')
		assert not match('abc', 'xyz')


def create_truncated_file(path):
	# Random text
	text = ''.join(random.choice('ABCDEFGHIJKLMNOPQRSTUVWXYZ') for _ in range(200))
	f = xopen(path, 'w')
	f.write(text)
	f.close()
	f = open(path, 'a')
	f.truncate(os.stat(path).st_size - 10)
	f.close()

# Disable these tests in Python 3.2 and 3.3
if not ((3, 2) <= sys.version_info[:2] <= (3, 3)):
	@raises(EOFError)
	def test_truncated_gz():
		with temporary_path('truncated.gz') as path:
			create_truncated_file(path)
			f = xopen(path, 'r')
			f.read()
			f.close()


	@raises(EOFError)
	def test_truncated_gz_iter():
		with temporary_path('truncated.gz') as path:
			create_truncated_file(path)
			f = xopen(path, 'r')
			for line in f:
				pass
			f.close()



# coding: utf-8
"""
Classes for writing and filtering of processed reads.

A Filter is a callable that has the read as its only argument. If it is called,
it returns True if the read should be filtered (discarded), and False if not.

To be used, a filter needs to be wrapped in one of the Wrapper classes. To determine
what happens to a read, a list of Wrappers with different filters is created and
each Wrapper is called in turn until one returns True. The main program will
determine whether and where to write the read(s) based on whether it was rejected
by a filter, and which one.
"""
from collections import OrderedDict

# Constants used when returning from a Filter’s __call__ method to improve
# readability (it is unintuitive that "return True" means "discard the read").
DISCARD = True
KEEP = False

class SingleWrapper(object):
    """
    This is for single-end reads and for paired-end reads, using the 'legacy' filtering mode
    (backwards compatibility). That is, if the first read matches the filtering criteria, the
    pair is discarded. The second read is not inspected.
    """
    def __init__(self, f):
        self.filtered = 0
        self.filter = f
        
    def __call__(self, read1, read2=None):
        if self.filter(read1):
            self.filtered += 1
            return DISCARD
        return KEEP

class PairedWrapper(object):
    """
    This is for paired-end reads, using the 'new-style' filtering where both reads are inspected.
    That is, the entire pair is discarded if at least 1 or 2 of the reads match the
    filtering criteria.
    """
    def __init__(self, f, min_affected=1):
        """
        min_affected -- values 1 and 2 are allowed.
            1 means: the pair is discarded if any read matches
            2 means: the pair is discarded if both reads match
        """
        if not min_affected in (1, 2):
            raise ValueError("min_affected must be 1 or 2")
        self.filtered = 0
        self.filter = f
        self.min_affected = min_affected

    def __call__(self, read1, read2):
        failures = 0
        if self.filter(read1):
            failures += 1
        if (self.min_affected - failures == 1) and self.filter(read2):
            failures += 1
        if failures >= self.min_affected:
            self.filtered += 1
            return DISCARD
        return KEEP

class FilterFactory(object):
    def __init__(self, paired, min_affected):
        self.paired = paired
        self.min_affected = min_affected
    
    def __call__(self, filter_type, *args, **kwargs):
        f = filter_type(*args, **kwargs)
        if self.paired == "both":
            return PairedWrapper(f, self.min_affected)
        else:
            return SingleWrapper(f)

class MergedReadFilter(object):
    def __call__(self, read):
        return read.merged

class TooShortReadFilter(object):
    def __init__(self, minimum_length):
        self.minimum_length = minimum_length
    
    def __call__(self, read):
        return len(read) < self.minimum_length

class TooLongReadFilter(object):
    def __init__(self, maximum_length):
        self.maximum_length = maximum_length
    
    def __call__(self, read):
        return len(read) > self.maximum_length

class NContentFilter(object):
    """
    Discards a reads that has a number of 'N's over a given threshold. It handles both raw
    counts of Ns as well as proportions. Note, for raw counts, it is a greater than comparison,
    so a cutoff of '1' will keep reads with a single N in it.
    """
    def __init__(self, count):
        """
        Count -- if it is below 1.0, it will be considered a proportion, and above and equal to
        1 will be considered as discarding reads with a number of N's greater than this cutoff.
        """
        assert count >= 0
        self.is_proportion = count < 1.0
        self.cutoff = count

    def __call__(self, read):
        """Return True when the read should be discarded"""
        n_count = read.sequence.lower().count('n')
        if self.is_proportion:
            if len(read) == 0:
                return False
            return n_count / len(read) > self.cutoff
        else:
            return n_count > self.cutoff

class UntrimmedFilter(object):
    """
    Return True if read is untrimmed.
    """
    def __call__(self, read):
        return read.match is None

class TrimmedFilter(object):
    """
    Return True if read is trimmed.
    """
    def __call__(self, read):
        return read.match is not None

class NoFilter(object):
    def __call__(self, read):
        return False

class Filters(object):
    def __init__(self, filter_factory):
        self.filters = OrderedDict()
        self.filter_factory = filter_factory
    
    def add_filter(self, filter_type, *args, **kwargs):
        self.filters[filter_type] = self.filter_factory(filter_type, *args, **kwargs)
    
    def filter(self, read1, read2=None):
        dest = NoFilter
        for filter_type, f in self.filters.items():
            if f(read1, read2):
                dest = filter_type
                # Stop writing as soon as one of the filters was successful.
                break
        return dest
    
    def __contains__(self, filter_type):
        return filter_type in self.filters
    
    def __getitem__(self, filter_type):
        return self.filters[filter_type]



# coding: utf-8
"""
Tests write output (should it return True or False or write)
"""
from __future__ import print_function, division, absolute_import

from atropos.filters import NContentFilter, DISCARD, KEEP, SingleWrapper, PairedWrapper
from atropos.seqio import Sequence

def test_ncontentfilter():
	# third parameter is True if read should be discarded
	params = [
		('AAA', 0, KEEP),
		('AAA', 1, KEEP),
		('AAACCTTGGN', 1, KEEP),
		('AAACNNNCTTGGN', 0.5, KEEP),
		('NNNNNN', 1, DISCARD),
		('ANAAAA', 1/6, KEEP),
		('ANAAAA', 0, DISCARD)
	]
	for seq, count, expected in params:
		filter = NContentFilter(count=count)
		_seq = Sequence('read1', seq, qualities='#'*len(seq))
		assert filter(_seq) == expected


def test_ncontentfilter_paired():
	params = [
		('AAA', 'AAA', 0, KEEP),
		('AAAN', 'AAA', 0, DISCARD),
		('AAA', 'AANA', 0, DISCARD),
		('ANAA', 'AANA', 1, KEEP),
	]
	for seq1, seq2, count, expected in params:
		filter = NContentFilter(count=count)
		filter_legacy = SingleWrapper(filter)
		filter_both = PairedWrapper(filter)
		read1 = Sequence('read1', seq1, qualities='#'*len(seq1))
		read2 = Sequence('read1', seq2, qualities='#'*len(seq2))
		assert filter_legacy(read1, read2) == filter(read1)
		# discard entire pair if one of the reads fulfills criteria
		assert filter_both(read1, read2) == expected



# coding: utf-8
from __future__ import print_function, division, absolute_import

from atropos.seqio import Sequence
from atropos.qualtrim import nextseq_trim_index

def test_nextseq_trim():
	s = Sequence('n', '', '')
	assert nextseq_trim_index(s, cutoff=22) == 0
	s = Sequence('n',
		'TCTCGTATGCCGTCTTATGCTTGAAAAAAAAAAGGGGGGGGGGGGGGGGGNNNNNNNNNNNGGNGG',
		'AA//EAEE//A6///E//A//EA/EEEEEEAEA//EEEEEEEEEEEEEEE###########EE#EA'
	)
	assert nextseq_trim_index(s, cutoff=22) == 33



# Parallel implementation of run_atropos. Works as follows:
#
# 1. Main thread creates N worker processes (where N is the number of threads to be allocated) and
# (optionally) one writer process.
# 2. Main thread loads batches of reads (or read pairs) from input file(s) and adds them to a queue
# (the input queue).
# 3. Worker processes take batches from the input queue, process them as atropos normally does,
# and either add the results to the result queue (if using a writer process) or write the results
# to disk. Each result is a dict mapping output file names to strings, where each string is a
# concatenation of reads (with appropriate line endings) to be written. A parameter also controls
# whether data compression is done by the workers or the writer.
# 4. If using a writer process, it takes results from the result queue and writes each string to
# its corresponding file.
# 5. When the main process finishes loading reads from the input file(s), it sends a signal to the
# worker processes that they should complete when the input queue is empty. It also singals the
# writer process how many total batches to expect, and the writer process exits after it has
# processed that many batches.
# 6. When a worker process completes, it adds a summary of its activity to the summary queue.
# 7. The main process reads summaries from the summary queue and merges them to create the complete
# summary, which is used to generate the report.
#
# There are several possible points of failure:
#
# 1. The main process may exit due to an unexpected error, or becuase the user forces it to exit
# (Ctrl-C). In this case, an attempt is made to cancel all processes before exiting.
# 2. A worker or writer process may exit due to an unknown error. To handle this, the main process
# checks that each process is alive whenver it times out writing to the input queue, and again when
# waiting for worker summaries. If a process has died, the program exits with an error since some data
# might have gotten lost.
# 3. More commonly, process will time out blocking on reading from or writing to a queue. Size
# limits are used (optionally) for the input and result queues to prevent using lots of memory. When
# few threads are allocated, it is most likely that the main and writer processes will block, whereas
# with many threads allocated the workers are most likely to block. Also, e.g. in a cluster
# environment, I/O latency may cause a "backup" resulting in frequent blocking of the main and workers
# processes. Finally, also e.g. in a cluster environment, processes may suspended for periods of time.
# Use of a hard timeout period, after which processes are forced to exit, is thus undesirable.
# Instead, parameters are provided for the user to tune the batch size and max queue sizes to their
# particular environment. Additionally, a "soft" timeout is used, after which log messages are
# escallated from DEBUG to ERROR level. The user can then make the decision of whether or not to kill
# the program.

from collections import defaultdict
import logging
from queue import Empty, Full
import time
from multiprocessing import Process, Queue, Value, cpu_count

from .report import *
from .seqio import Writers, FormatError
from .xopen import get_compressor

__author__ = "John Didion"

# Max time to wait between retrying operations
RETRY_INTERVAL = 5

def wait_on(condition, message="Waiting {}", timeout=None, fail_callback=None, wait=None, timeout_callback=None):
    if wait is True:
        wait = lambda: time.sleep(RETRY_INTERVAL)
    elif isinstance(wait, int):
        wait_time = wait
        wait = lambda: time.sleep(wait_time)
    wait_start = None
    while True:
        result = condition()
        if result is not False:
            return result
        if fail_callback:
            fail_callback()
        now = time.time()
        if not wait_start:
            wait_start = now
        else:
            waiting = now - wait_start
            msg = message.format("for {} seconds".format(round(waiting, 1)))
            if timeout is not None and waiting >= timeout:
                logging.getLogger().error(msg)
                if timeout_callback:
                    if issubclass(timeout_callback, Exception):
                        raise timeout_callback()
                    else:
                        timeout_callback()
            else:
                logging.getLogger().debug(msg)
            if wait:
                wait()

def enqueue(queue, item, message="Waiting to enqueue item {}", block_timeout=RETRY_INTERVAL, **kwargs):
    def condition():
        try:
            queue.put(item, block=True, timeout=block_timeout)
            return True
        except Full:
            return False
    wait_on(condition, message=message, **kwargs)

def dequeue(queue, message="Waiting to dequeue item {}", block_timeout=RETRY_INTERVAL, **kwargs):
    def condition():
        try:
            return queue.get(block=True, timeout=block_timeout)
        except Empty:
            return False
    return wait_on(condition, message=message, **kwargs)

class Control(object):
    def __init__(self, initial_value):
        self.control = Value('l', initial_value)

    def check_value(self, value, lock=False):
        return self.get_value(lock=lock) == value

    def check_value_positive(self, lock=False):
        return self.get_value(lock=lock) > 0

    def get_value(self, lock=True):
        if lock:
            with self.control.get_lock():
                return self.control.value
        else:
            return self.control.value

    def set_value(self, value):
        with self.control.get_lock():
            self.control.value = value

class WorkerProcess(Process):
    def __init__(self, index, modifiers, filters, formatters, input_queue,
                 result_handler, summary_queue, timeout):
        """
        input_queue: queue with batches of records to process
        output_queue: queue where results are written
        summary_queue: queue where summary information is written
        control: shared value with value 0 unless the thread
        should die (< 0) or there are no more batches coming (> 0).
        """
        super(WorkerProcess, self).__init__(name="Worker process {}".format(index))
        self.index = index
        self.modifiers = modifiers
        self.filters = filters
        self.formatters = formatters
        self.input_queue = input_queue
        self.result_handler = result_handler
        self.summary_queue = summary_queue
        self.timeout = timeout
        self.processed_reads = 0
        self.total_bp1 = 0
        self.total_bp2 = 0
        self.seen_batches = set()
    
    def run(self):
        def iter_batches():
            while True:
                batch = dequeue(
                    self.input_queue,
                    message="{} waiting on batch {{}}".format(self.name),
                    timeout=self.timeout)
                yield batch
        
        def enqueue_summary(error=False):
            if error:
                process_stats = adapter_stats = None
            else:
                process_stats = collect_process_statistics(
                    self.processed_reads, self.total_bp1, self.total_bp2,
                    self.modifiers, self.filters, self.formatters)
                adapter_stats = summarize_adapters(self.modifiers.get_modifier_pair(AdapterCutter))
            enqueue(
                self.summary_queue,
                (self.index, self.seen_batches, process_stats, adapter_stats),
                "{} waiting to queue summary {{}}".format(self.name),
                timeout=self.timeout
            )
        
        try:
            self.result_handler.start(self)
            
            for batch in iter_batches():
                if batch is None:
                    break
                
                batch_num, (batch_size, records) = batch
                self.seen_batches.add(batch_num)
                self.processed_reads += batch_size
                
                result = defaultdict(lambda: [])
                for record in records:
                    reads, bp = self.modifiers.modify(record)
                    self.total_bp1 += bp[0]
                    self.total_bp2 += bp[1]
                    dest = self.filters.filter(*reads)
                    self.formatters.format(result, dest, *reads)
                
                self.result_handler.write_result(batch_num, result)
            
            logging.getLogger().debug("{} finished; processed {} batches, {} reads ({},{} bp)".format(
                self.name, len(self.seen_batches), self.processed_reads, self.total_bp1, self.total_bp2))
            
            logging.getLogger().debug("{} sending summary".format(self.name))
            enqueue_summary()
            
            logging.getLogger().debug("{} exiting normally".format(self.name))
        except:
            logging.getLogger().error("Unexpected error in {}".format(self.name), exc_info=True)
            enqueue_summary(error=True)
        finally:
            self.result_handler.finish()

# Control values
# Writer is running normally
WRITER_ACTIVE = 0
# Writer should exit (usu. because there were no reads in the user input)
WRITER_ERROR = -1

class WriterProcess(Process):
    """
    Thread that accepts results from the worker threads
    and writes the results to output file(s). Each batch
    is expected to be (batch_num, path, records), where
    path is the destination file and records is a string.
    Not guaranteed to preserve the original order of
    sequence records.

    writers: Writers object, which manages opening and writing
    to files.
    queue: input queue.
    control: a shared value for communcation with the main
    process.
    timeout: seconds to wait for next batch before complaining
    """
    def __init__(self, result_handler, queue, control, timeout=60):
        super(WriterProcess, self).__init__(name="Writer process")
        self.result_handler = result_handler
        self.queue = queue
        self.control = control
        self.timeout = timeout
        self.seen_batches = set()
        self.num_batches = None

    def run(self):
        class Done(Exception): pass
        class Killed(Exception): pass

        def fail_callback():
            if self.num_batches is None and self.control.check_value_positive():
                self.num_batches = self.control.get_value()
            if self.num_batches is not None and len(self.seen_batches) >= self.num_batches:
                raise Done()

        def timeout_callback():
            if self.num_batches is not None:
                missing = set(range(1, self.num_batches+1)) - self.seen_batches
                logging.getLogger().error("Writer thread still missing batches {} of {}".format(
                    ",".join(str(i) for i in missing), self.num_batches))

        def iter_batches():
            while True:
                batch = dequeue(
                    self.queue,
                    message="Writer waiting on result {}",
                    timeout=self.timeout,
                    fail_callback=fail_callback,
                    timeout_callback=timeout_callback)
                yield batch

        try:
            self.result_handler.start(self)
            
            for batch_num, result in iter_batches():
                self.seen_batches.add(batch_num)
                self.result_handler.write_result(batch_num, result)
        except Done:
            logging.getLogger().debug("Writer process exiting normally")
        except Killed:
            logging.getLogger().debug("Writer process exited early")
        except:
            logging.getLogger().error("Unexpected error in writer process", exc_info=True)
            self.control.set_value(WRITER_ERROR)
        finally:
            num_batches = self.control.get_value(lock=True)
            self.result_handler.finish(num_batches if num_batches > 0 else None)

class WorkerResultHandler(object):
    """
    Wraps a ResultHandler and compresses results prior
    to writing.
    """
    def __init__(self, handler):
        self.handler = handler

    def start(self, worker):
        self.handler.start(worker)
    
    def write_result(self, batch_num, result):
        """
        Give a dict mapping files to lists of strings,
        join the strings and compress them (if necessary)
        and then return the property formatted result
        dict.
        """
        self.handler.write_result(batch_num, dict(self.prepare_file(*item)
            for item in result.items()))
    
    def prepare_file(self, path, strings):
        return (path, "".join(strings))
    
    def finish(self, total_batches=None):
        self.handler.finish()

class CompressingWorkerResultHandler(WorkerResultHandler):
    """
    Wraps a ResultHandler and compresses results prior
    to writing.
    """
    def start(self, worker):
        super(CompressingWorkerResultHandler, self).start(worker)
        self.file_compressors = {}
    
    def prepare_file(self, path, strings):
        compressor = self.get_compressor(path)
        if compressor:
            return ((path, 'wb'), compressor.compress(b''.join(
                s.encode() for s in strings)))
        else:
            return ((path, 'wt'), "".join(strings))
    
    def get_compressor(self, filename):
        if filename not in self.file_compressors:
            self.file_compressors[filename] = get_compressor(filename)
        return self.file_compressors[filename]

class QueueResultHandler(object):
    """
    ResultHandler that writes results to the output queue.
    """
    def __init__(self, queue):
        self.queue = queue

    def start(self, worker):
        self.message = "{} waiting to queue result {{}}".format(worker.name)
        self.timeout = worker.timeout

    def write_result(self, batch_num, result):
        enqueue(
            self.queue,
            (batch_num, result),
            self.message,
            timeout=self.timeout
        )

    def finish(self, total_batches=None):
        pass

class WriterResultHandler(object):
    """
    ResultHandler that writes results to disk.
    """
    def __init__(self, writers, compressed=False, use_suffix=False):
        self.writers = writers
        self.compressed = compressed
        self.use_suffix = use_suffix

    def start(self, worker):
        if self.use_suffix:
            self.writers.suffix = ".{}".format(worker.index)

    def write_result(self, batch_num, result):
        self.writers.write_result(result, self.compressed)

    def finish(self, total_batches=None):
        self.writers.close()

class OrderPreservingWriterResultHandler(WriterResultHandler):
    """
    Writer thread that is less time/memory efficient, but is
    guaranteed to preserve the original order of records.
    """
    def start(self, worker):
        super(OrderPreservingWriterResultHandler, self).__init__(worker)
        self.pending = PendingQueue()
        self.cur_batch = 1

    def write_result(self, batch_num, result):
        if batch_num == self.cur_batch:
            self.writers.write_result(result, self.compressed)
            self.cur_batch += 1
            self.consume_pending()
        else:
            self.pending.push(batch_num, result)

    def finish(self, total_batches):
        if total_batches is not None:
            self.consume_pending()
            if self.cur_batch != total_batches:
                raise Exception("OrderPreservingWriterResultHandler finishing without having seen "
                                "{} batches".format(total_batches))
        self.writers.close()

    def consume_pending(self):
        while (not self.pending.empty) and (self.cur_batch == pending.min_priority):
            self.writers.write_result(pending.pop(), self.compressed)
            self.cur_batch += 1

class PendingQueue(object):
    def _init(self, max_size=None):
        self.queue = {}
        self.max_size = max_size
        self.min_priority = None

    def push(self, priority, value):
        assert not self.full
        self.queue[priority] = value
        if self.min_priority is None or priority < self.min_priority:
            self.min_priority = priority

    def pop(self):
        assert not self.empty
        value = self.queue.pop(self.min_priority)
        if self.empty:
            self.min_priority = None
        else:
            self.min_priority = min(self.queue.keys())
        return value

    @property
    def full(self):
        return self.max_size and len(self.queue) >= self.max_size

    @property
    def empty(self):
        return len(self.queue) == 0

def run_parallel(reader, modifiers, filters, formatters, writers, threads=1, timeout=30,
                 preserve_order=False, input_queue_size=0, result_queue_size=0,
                 use_writer_process=True, worker_compression=False):
    """
    Execute atropos in parallel mode.

    reader 				:: iterator over batches of reads (most likely a BatchIterator)
    modifiers 			::
    filters 			::
    formatters 			::
    writers				::
    threads				:: number of worker threads to use; additional threads are used
                        for the main proccess and the writer process (if requested).
    timeout				:: number of seconds after which waiting processes escalate their
                        messages from DEBUG to ERROR.
    preserve_order 		:: whether to preserve the input order of reads when writing
                        (only valid when `use_writer_process=True`)
    input_queue_size 	:: max number of items that can be in the input queue, or 0 for
                        no limit (be warned that this could explode memory usage)
    result_queue_size	:: max number of items that can be in the result queue, or 0 for
                        no limit (be warned that this could explode memory usage)
    use_writer_process	:: if True, a separate thread will be used to write results to
                        disk. Otherwise, each worker thread will write its results to
                        an output file with a '.N' extension, where N is the thread index.
                        This is useful in cases where the I/O is the main bottleneck.
    worker_compression	If True, the worker processes perform data compression, otherwise
                        the writer process performs compression.
    """
    logging.getLogger().debug(
        "Starting atropos in parallel mode with threads={}, timeout={}".format(threads, timeout))

    timeout = max(timeout, RETRY_INTERVAL)

    # Queue by which batches of reads are sent to worker threads
    input_queue = Queue(input_queue_size)
    # Queue by which results are sent from the worker threads to the writer thread
    result_queue = Queue(result_queue_size)
    # Queue for threads to send summary information back to main process
    summary_queue = Queue(threads)
    # Aggregate summary
    summary = Summary(trimmer_classes=modifiers.get_trimmer_classes())
    # Return code (0=normal, anything else is an error)
    rc = 0

    if use_writer_process:
        worker_result_handler = QueueResultHandler(result_queue)
        if worker_compression:
            worker_result_handler = CompressingWorkerResultHandler(worker_result_handler)
        else:
            worker_result_handler = WorkerResultHandler(worker_result_handler)
        
        # Shared variable for communicating with writer thread
        writer_control = Control(WRITER_ACTIVE)
        # result handler
        if preserve_order:
            writer_result_handler = OrderPreservingWriterResultHandler(
                writers, compressed=worker_compression)
        else:
            writer_result_handler = WriterResultHandler(
                writers, compressed=worker_compression)
        # writer process
        writer_process = WriterProcess(writer_result_handler, result_queue, writer_control, timeout)
        writer_process.start()
    else:
        worker_result_handler = WorkerResultHandler(WriterResultHandler(writers, use_suffix=True))

    # worker processes
    worker_processes = [
        WorkerProcess(
            i, modifiers, filters, formatters, input_queue,
            worker_result_handler, summary_queue, timeout)
        for i in range(threads)
    ]
    # start workers
    for worker in worker_processes: worker.start()

    def ensure_alive():
        alive = [worker.is_alive() for worker in worker_processes]
        if not all(alive):
            raise Exception("One or more worker process exited: {}".format(",".join(
                str(i) for i in range(len(alive)) if not alive[i])))
        if use_writer_process and not (writer_process.is_alive() and writer_control.check_value(WRITER_ACTIVE)):
            raise Exception("Writer process exited")

    def enqueue_all(iterable):
        num_items = 0
        for item in iterable:
            def condition():
                try:
                    input_queue.put(item, block=True, timeout=RETRY_INTERVAL)
                    return True
                except Full:
                    return False
            wait_on(
                condition,
                message="Main process waiting to queue item {}",
                timeout=timeout,
                fail_callback=ensure_alive)
            num_items += 1
        return num_items

    def wait_on_process(process, terminate=False):
        timeout_callback = lambda: process.terminate() if terminate else None
        wait_on(
            lambda: not process.is_alive(),
            message="Waiting on {} to terminate {{}}".format(process.name),
            timeout=timeout,
            wait=lambda: process.join(RETRY_INTERVAL),
            timeout_callback=timeout_callback)

    try:
        # Add batches of reads to the input queue. Provide a timeout callback
        # to check that subprocesses are alive.
        num_batches = enqueue_all(enumerate(reader, 1))
        logging.getLogger().debug(
            "Main loop complete; saw {} batches".format(num_batches))

        # Tell the worker processes no more input is coming
        enqueue_all((None,) * threads)

        # Tell the writer thread the max number of batches to expect
        if use_writer_process:
            writer_control.set_value(num_batches)

        # Wait for workers to exit
        #for worker in worker_processes:
        #	# Wait for worker to exit
        #	wait_on_process(worker)

        # Wait for all summaries to be available on queue
        def summary_timeout_callback():
            alive = [worker.is_alive() for worker in worker_processes]
            if any(alive):
                missing = ",".join(str(i) for i in range(len(alive)) if alive[i])
                logging.getLogger().error(
                    "Workers are still alive and haven't returned summaries: {}".format(missing))
        wait_on(
            lambda: summary_queue.full(),
            "Waiting on worker summaries {}",
            timeout=timeout,
            wait=True,
            timeout_callback=summary_timeout_callback)

        # Process summary information from worker threads
        logging.getLogger().debug("Processing summary information from worker threads")
        seen_summaries = set()
        seen_batches = set()
        def summary_fail_callback():
            missing_summaries = set(range(1, threads)) - seen_summaries
            raise Exception("Missing summaries from processes {}".format(
                ",".join(str(s) for s in missing)))
        for i in range(1, threads+1):
            batch = dequeue(
                summary_queue,
                fail_callback=summary_fail_callback)
            worker_index, worker_batches, process_stats, adapter_stats = batch
            if process_stats is None or adapter_stats is None:
                raise Exception("Worker process {} died unexpectedly".format(worker_index))
            else:
                logging.getLogger().debug("Processing summary for worker {}".format(worker_index))
            seen_summaries.add(worker_index)
            seen_batches |= worker_batches
            summary.add_process_stats(process_stats)
            summary.add_adapter_stats(adapter_stats)

        # Check if any batches were missed
        if num_batches > 0:
            missing_batches = set(range(1, num_batches+1)) - seen_batches
            if len(missing_batches) > 0:
                raise Exception("Workers did not process batches {}".format(
                    ",".join(str(b) for b in missing_batches)))

        if use_writer_process:
            # Wait for writer to complete
            wait_on_process(writer_process)

    except KeyboardInterrupt as e:
        logging.getLogger().error("Interrupted")
        rc = 130

    except IOError as e:
        if e.errno == errno.EPIPE:
            rc = 1
        else:
            raise

    except (FormatError, EOFError) as e:
        logging.getLogger().error("Cutadapt error", exc_info=True)
        rc = 1

    except Exception as e:
        logging.getLogger().error("Unknown error", exc_info=True)
        rc = 1

    finally:
        logging.getLogger().debug("Waiting for reader to close")
        reader.close()

        # notify all threads that they should stop
        logging.getLogger().debug("Exiting all processes")
        def kill(process):
            if rc <= 1:
                wait_on_process(process, terminate=True)
            elif process.is_alive():
                process.terminate()
        for process in worker_processes:
            kill(process)
        if use_writer_process:
            kill(writer_process)
    
    return (rc, summary.finish())
